/**
 * The native methods - encryption primitives.
 * @file gproxy/CFuncts.java
 **/
package gproxy;

/**
 * C method interface. Generate .h file with javah:
 * <code>
 *  javah -classpath .. -o ../include/CFuncts.h gproxy.CFuncts
 * </code>
 * @author Christian Grothoff
 **/
public class CFuncts {

    static {
	System.loadLibrary("gproxy");
    }

    /**
     * Compute CRC32 value of the given data array.
     * @param data the array to compute the CRC32 of
     * @return the crc32 value
     **/
    public static int crc32(byte[] data) {
	return crcNative(data);
    }
     
    /**
     * Compute the CRC32 checksum of the given data array.
     * @param data the array to compute the CRC32 of
     * @return the crc32 value
     **/
    private native static int crcNative(byte[] data); 
    
    /**
     * Compute the RIPE160 hash code of the given data array.
     * @param data the block to compute the hash of
     * @return the hash value (160 bit)
     **/
    public static byte[] hash(byte[] data) {
	return hashNative(data);
    }
    
    /**
     * Compute the RIPE160 hash code of the given data array.
     * @param data the block to compute the hash of
     * @return the hash value (160 bit)
     **/
    private native static byte[] hashNative(byte[] data); 

    /**
     * Decrypt (blowfish) the given data array computing the
     * key and initialization vector from the given 20-byte 
     * hashcode.
     * @param key the key to decrypt (including IV), must be 20 bytes
     * @param data the data to decrypt, must be 1k
     * @return the plaintext, will be 1k
     **/
    public static byte[] decrypt(byte[] key, 
				 byte[] data) {
	/*
	System.out.println("Decrypting with key " + LoopbackBuildThread.bytesToHex(key) + ": ");
	for (int i=0;i<data.length;i++)
	    System.out.print((int) (((char)data[i])&255) + "-" + i + " ");
	*/
	if (key.length != Constants.HASH_SIZE)
	    throw new Error("bad keylength");
	if (data.length != Constants.CONTENT_BLOCK_SIZE)
	    throw new Error("bad datalength");
	byte[] result = decryptNative(key, data);
	/*
	  System.out.println("\nDecrypted: ");
	  for (int i=0;i<result.length;i++)
	  System.out.print((int) (((char)result[i])&255) + "-" + i + " ");
	  System.out.println("\nDecrypted (hex): ");
	  System.out.println(LoopbackBuildThread.bytesToHex(result));
	*/
	return result;
    }

    /**
     * Get the decrypted version of the input
     * @param key the key to decrypt (including IV), must be 20 bytes
     * @param data the data to decrypt, must be 1k
     * @return the plaintext, will be 1k
     **/
    private native static byte[] decryptNative(byte[] key, 
					       byte[] data); 
    
    /**
     * Sleep for n seconds.
     * @param n the number of seconds
     **/
    public static void sleep(int n) {
	Object o = new Object();
	synchronized (o) {
	    try {
		o.wait(1000*n);  
	    } catch (InterruptedException ie) {
		throw new Error("Unexpected interrupted exception caught!");
	    }
	} 
    }

    /**
     * This function compares the hash code of data that has arrived with the hash code of the
     * data this node is expecting.
     * @param h1 hash code of incoming data
     * @param h2 hash code this node is expecting
     * @return boolean true if the hashes are equal, false if not
     **/
    public static boolean hashEquals(byte[] h1, byte[] h2) {
	for (int i=0;i<h1.length;i++)
	    if (h1[i] != h2[i])
		return false;
	return true;
    } 


    /**
     * This function converts a byte-array to a hexadecimal string.
     * See hash2hex().
     * @param bytes byte array to be converted
     * @return hex String representing array
     **/

    public static String bytes2hexString(byte[] bytes) {
        int this_nibble;

        StringBuffer ret = new StringBuffer(); 
 
        for (int i = 0; i < bytes.length; i++) {
            this_nibble = (bytes[i] & (byte)(0xF0)) >> 4;
            ret.append(Integer.toString(this_nibble, 16).toUpperCase());
            this_nibble = bytes[i] >> 4;              
            ret.append(Integer.toString(this_nibble, 16).toUpperCase());
        }
        return ret.toString();
    }
} // end of CFuncts
