/** 
 * Converter converts arrays of bytes to primitives and back. 
 */
package gproxy;

import java.io.IOException;

/**
 * Simple Converter for primitives to byte array.
 * @author James Liang 
 * @author Christian Grothoff
 */
public final class Converter {

    // can't be instantiated
    private Converter() {
    }

    public static final int LITTLE_ENDIAN = 0;
    public static final int BIG_ENDIAN = 1;

    public static final int TARGETENDIAN = LITTLE_ENDIAN;
    public static final int CURRENTENDIAN = BIG_ENDIAN;
	
    /**
     * Reads a long from data 
     * @param data the byte array to read from
     * @param location the place to start reading.
     * @return the char stored in byte array
     **/
    public static final long getLong (byte[] data, int location) {
	return (long)
	    ( ( ((((long)data[location  ])      )      ) << 56) 
	    + ( ((((long)data[location+1])&0xff)) << 48)
	    + ( ((((long)data[location+2])&0xff)) << 40) 
	    + ( ((((long)data[location+3])&0xff)) << 32)
	    + ( ((((long)data[location+4])&0xff)) << 24) 
	    + ( ((((long)data[location+5])&0xff)) << 16)
	    + ( ((((long)data[location+6])&0xff)) <<  8) 
	    + ( ((((long)data[location+7])&0xff))      ));
    }

    /**
     * Writes a long from data 
     * @param data the byte array to store to.
     * @param location the place to start writing.
     * @param value the value to write
     **/
    public static final void setLong (byte[] data, int location, long value) {
	data[location  ]=(byte)((value>>56) );
	data[location+1]=(byte)((value>>48) & 0xFF);
	data[location+2]=(byte)((value>>40) & 0xFF);
	data[location+3]=(byte)((value>>32) & 0XFF);
	data[location+4]=(byte)((value>>24) & 0xFF);
	data[location+5]=(byte)((value>>16) & 0xFF);
	data[location+6]=(byte)((value>> 8) & 0xFF);
	data[location+7]=(byte)((value    ) & 0XFF);
    }
    
    /**
     * Reads a short from data 
     * @param data the byte array to read from
     * @param location the place to start reading.
     * @return the char stored in byte array
     **/
    public static final short getShort (byte[] data, int location) {
	return (short)
	    ( ( ((((short)data[location+0])&0xff)) <<  8) 
	    + ( ((((short)data[location+1])&0xff))      ));
    }

    /**
     * Writes a short from data 
     * @param data the byte array to store to.
     * @param location the place to start writing.
     * @param value the value to write
     **/
    public static final void setShort (byte[] data, int location, short value) {
	data[location  ]=(byte)((value>> 8) & 0x80FF);
	data[location+1]=(byte)((value    ) & 0XFF);
     }
    
    /**
     * Reads an int from data 
     * @param data the byte array to read from
     * @param location the place to start reading.
     * @return the char stored in byte array
     **/
    public static final int getInt (byte[] data, int location) {
	return   
	    ( ( ((((int)data[location  ])      )      ) << 24) 
	    + ( ((((int)data[location+1])&0xff)) << 16)
	    + ( ((((int)data[location+2])&0xff)) <<  8) 
	    + ( ((((int)data[location+3])&0xff))      ));
    }

    /**
     * Writes an int from data 
     * @param data the byte array to store to.
     * @param location the place to start writing.
     **/
    public static final void setInt (byte[] data, int location, int value) {
	data[location  ]=(byte)((value>>24)       );
	data[location+1]=(byte)((value>>16) & 0xFF);
	data[location+2]=(byte)((value>> 8) & 0xFF);
	data[location+3]=(byte)((value    ) & 0XFF);
    }
    
    /**
     * Reads a char from data 
     * @param data the byte array to read from
     * @param location the place to start reading.
     * @return the char stored in byte array
     **/
    public static final char getChar(byte[] data,int location) {
	return (char) (((int)(data[location]   - 256 * (data[location] >> 7)) << 8) + 
		       ((int)(data[location+1] - 256 * (data[location+1] >> 7))));
    }
    
    /**
     * Writes a char from data 
     * @param data the data array to store to.
     * @param location the place to start writing.
     **/
    public static final void setChar(byte[] data,int location,char value) {
	data[location]   = (byte) (value >> 8);
	data[location+1] = (byte) (value & 0xFF);
    }

    /**
     * Converts a value 2-byte from little endian to big endian or
     * from big endian to little endian
     *
     * @param 2-byte value to be converted
     * @return the same number converted.
     **/
    public static final short reverseEndian(short value) {	
	byte [] tmpBlock = new byte[2];
	byte tmp;
	setShort(tmpBlock,0,value);
	tmp = tmpBlock[0];
	tmpBlock[0] = tmpBlock[1];
	tmpBlock[1] = tmp;
	return getShort(tmpBlock,0);
    }

    /**
     * Converts a value 2-byte from little endian to big endian or
     * from big endian to little endian
     *
     * @param 2-byte value to be converted
     * @return the same number converted.
     **/
    public static final char reverseEndian(char value) {	
	byte [] tmpBlock = new byte[2];
	byte tmp;
	setChar(tmpBlock,0,value);
	tmp = tmpBlock[0];
	tmpBlock[0] = tmpBlock[1];
	tmpBlock[1] = tmp;
	return getChar(tmpBlock,0);
    }

    /**
     * Converts a value 4-byte from little endian to big endian or
     * from big endian to little endian
     *
     * @param 4-byte value to be converted
     * @return the same number converted.
     **/
    public static final int reverseEndian(int value) {	
	byte [] tmpBlock = new byte[4];
	byte [] retVal = new byte[4];
	setInt(tmpBlock,0,value);
	retVal[3] = tmpBlock[0];
	retVal[2] = tmpBlock[1];
	retVal[1] = tmpBlock[2];
	retVal[0] = tmpBlock[3];
	return getInt(retVal,0);
    }

    /**
     * Converts a value 8-byte from little endian to big endian or
     * from big endian to little endian
     *
     * @param 8-byte value to be converted
     * @return the same number converted.
     **/
    public static final long reverseEndian(long value) {	
	byte [] tmpBlock = new byte[8];
	byte [] retVal = new byte[8];
	setLong(tmpBlock,0,value);
	for (int i =0;i<8;i++)
	    retVal[7-i] = tmpBlock[i];
	return getLong(retVal,0);
    }

    /** 
     * reads an UNSIGNED int from the byteCode (yepp, ignore the sign!)
     * @param offset the offset of the returned int in the bytecode 
     * @return the int stored at offset
     **/    
    public static int readUnsignedInt(byte[] data, int offset) { 
	return 
	    (int) (getChar (data,offset  ) << 16) + 
	    (int) (getChar (data,offset+2)      ); 
    }

    /** 
     * reads an UNSIGNED int from the byteCode (yepp, ignore the sign!)
     * and return it as a long
     * @param offset the offset of the returned int in the bytecode 
     * @return the int stored at offset
     **/    
    public static final long readUnsignedIntAsLong(byte[] data, int offset) { 
	return 
	    (long) (getChar (data,offset  ) << 16) + 
	    (long) (getChar (data,offset+2)      ); 
    }

    public static final char convertEndian(char value) {
	if ( TARGETENDIAN != CURRENTENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final short convertEndian(short value) {
	if ( TARGETENDIAN != CURRENTENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final int convertEndian(int value) {
	if ( TARGETENDIAN != CURRENTENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final long convertEndian(long value) {
	if ( TARGETENDIAN != CURRENTENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final char convertBigEndian(char value) {
	if ( CURRENTENDIAN != BIG_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final short convertBigEndian(short value) {
	if ( CURRENTENDIAN != BIG_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final int convertBigEndian(int value) {
	if ( CURRENTENDIAN != BIG_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final long convertBigEndian(long value) {
	if ( CURRENTENDIAN != BIG_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final char convertLittleEndian(char value) {
	if ( CURRENTENDIAN != LITTLE_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final short convertLittleEndian(short value) {
	if ( CURRENTENDIAN != LITTLE_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final int convertLittleEndian(int value) {
	if ( CURRENTENDIAN != LITTLE_ENDIAN )
	    return reverseEndian(value);
	return value;
    }

    public static final long convertLittleEndian(long value) {
	if ( CURRENTENDIAN != LITTLE_ENDIAN )
	    return reverseEndian(value);
	return value;
    }


} // end of Converter







