/**
 * Parses a query into its and and not portions - note that we do
 * not allow parenthesization.
 * @file gproxy/QueryParser.java
 **/
package gproxy;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class takes a set of strings retrieved from the commandline
 * and divides them up into included and excluded terms for queries.
 * It does not do complex boolean expressions!!!! The inclusion and
 * exclusion is on a PER TERM basis.
 * @author Krista Bennett
 * @author Christian Grothoff
 **/
public class QueryParser {

    /**
     * The actual query string submitted by the user
     **/
    private final String query_;

    /**
     * QueryModel that will be used to handle this query set
     **/
    private final QueryModel qm_;

    /**
     * Keywords of included queries
     **/
    private final Vector and_queries_ = new Vector();

    /**
     * Keywords of excluded queries
     **/
    private final Vector not_queries_ = new Vector();

    /**
     * List of all subqueries from within this query, in order
     **/
    private final Vector query_list_ = new Vector();

    /**
     * Object used to store/evaluate returned files' satisfaction
     * of the query string.
     **/
    private final QMatches q_match_;

    /**
     * Used to keep track of root inodes we've already displayed via
     * GProxy (these have been found complete at SOME time - may not still be)
     **/
    private Hashtable displayed_roots_;

    /**
     * String representing the AND operator
     **/
    private static final String and_value_ = new String("AND");

    /**
     * String representing the NOT operator
     **/
    private static final String not_value_ = new String("NOT");

    /**
     * Length of AND and NOT operator strings.
     **/
    private static final int and_not_len_ = 3;

    /**
     * Query processing index. Only has meaning during parsing.
     **/
    private int raw_index_ = 0;

    /**
     * Query processing value - is the next term inclusive? 
     * Only has meaning during parsing.
     **/
    private boolean is_and_ = true;

    /**
     * The constructor. Get the list of strings from the command line.
     * @param query - array of strings, one per "word" from
     *        the commandline.
     * @param qm the QueryModel to notify if a file matching
     *        the query has been found
     **/
    public QueryParser(String query,
		       QueryModel qm) {
	this.query_ = query;
	this.qm_ = qm;
        this.q_match_ = new QMatches();
        this.displayed_roots_ = new Hashtable();

        parseInternal(); // Parse away.

        if (GProxy.DEBUG) 
            printResults(); 
    }     


    /**
     * This method is an internal method that does the actual parsing
     * of the string, separating out query terms from operators and 
     * placing query terms both in their corresponding "and" and "not"
     * categories, as well as creating an in-order list of queries
     * contained within the string received by the constructor.
     **/
    private void parseInternal() {

        StringBuffer current_token = new StringBuffer(); /* current subquery */

        while (raw_index_ < query_.length()) {
            current_token.append(query_.charAt(raw_index_++));
 
            /* see if this is the end of the term and an AND or NOT follows */
            if (raw_index_ < query_.length() && 
                query_.charAt(raw_index_) == ' ') {
 
                /* see if this is an actual AND or NOT operator */
                if ((raw_index_ + and_not_len_) < query_.length()) { 

                    /* if this isn't an AND, is it a NOT? */
                    if (!checkOp(true, current_token))  {
                        if (checkOp(false, current_token)) {
                            current_token = new StringBuffer(); 
                        }
                    }
                    else
                        current_token = new StringBuffer();

                    /* it's not the end of a term. Keep going. */
                } 
            } 
        } 

        /* Commit last token if we haven't already */
        if (current_token.length() != 0)
            commitToken(current_token);
    } 


    /**
     * Process the current subquery token - put it into the main
     * query list, as well as the appropriate list corresponding
     * to its operator.
     * @param token The subquery string to be committed
     **/
    private void commitToken(StringBuffer token) {

        /* Decide whether to put this into the AND or NOT list */
        if (is_and_) {
           /* pop this into the AND list */
           and_queries_.addElement(token.toString());	
        }
        else {
           /* pop this into the NOT list */
           not_queries_.addElement(token.toString());
        }

        /* Put this into the overall list */
        query_list_.addElement(token.toString());
    }


    /**
     * This checks to see if the substring at the given index of
     * length oplen matches the operator string passed in.
     * @param opstring AND or NOT
     * @param oplen length of operator string passed in
     * @param index start of substring to check
     * @return true if we match the operator, false if not
     **/
    private boolean opMatch(String opstring, int oplen, int index) {
        return opstring.regionMatches(0, query_, index, oplen);
    }
 

    /**
     * Checks to see if this AND or NOT is part of a subquery instead 
     * of being a standalone operator.
     * @param index string index of 
     * @return true is this is a part of a query substring,
     *         false otherwise.
     **/ 
    private boolean isSearchPart(int index) {
        if (index + 1 < query_.length()) {
            /* is this an operator token or part of a search? */
            if (query_.charAt(index + 1) != ' ') {
                return true;  
            } 
        }
        return false;
    }


    /**
     * This checks to see whether the next term in the string,
     * indicated by the index submitted, matches the operator
     * indicated by the check_and flag. 
     *
     * If it does, the current token is "finished off" and committed, 
     * the flag which indicates the kind of operator (is_and_) for the next
     * term is set, and the new index into the raw string is returned
     * via the new_index Integer IF we have found an operator so
     * that the operator characters are not processed again. 
     *
     * Note that if the AND or NOT are part of a substring (e.g. ANDI or
     * NOTT), they will not be processed as operators but left as
     * normal search terms.
     *
     * @param check_and true if checking for AND, false if for NOT
     * @param token the current subquery being processed by caller
     * @return true if indicated operator found, false if not
     **/
    private boolean checkOp(boolean check_and, StringBuffer token) {
        boolean op_matched;

        if (check_and)
            op_matched = opMatch(and_value_, and_not_len_, raw_index_ + 1);
        else
            op_matched = opMatch(not_value_, and_not_len_, raw_index_ + 1);

        if (op_matched) {
            /* Make sure this isn't just part of a search term */
            if (isSearchPart(raw_index_ + and_not_len_))
                return false;
 
            /* It's for real - end this token and commit it. */
            commitToken(token);
 
            /* Indicate that the next one is to be ANDed or NOTed. */
            is_and_ = check_and;

            /* advance the index */
            raw_index_ += and_not_len_ + 2; // +2 for spaces
            return true;
        }
        else
           return false;
    }

    /**
     * @return the priority of this request
     **/
    public int getPriority() {
	return 0;
    }

    /**
     * Obtain the list of keywords that we are
     * looking for.
     * @return an array of the keywords
     **/
    public String[] getQueries() {
	String[] ret = new String[query_list_.size()];

        int count = 0;
        for (Enumeration e = query_list_.elements(); e.hasMoreElements(); ) 
            ret[count++] = (String)(e.nextElement());

	return ret; 
    }

    /**
     * Take notice that a new root-node matching one of the keywords has
     * been found. This must be remembered (in the future, AND queries
     * may become true due to this). If this already completes an
     * "AND" set for this query (and the "NOT" set is still empty),
     * return true, otherwise false. The QueryParser must
     * also initiate displaying of the node using the
     * QueryModel.
     * @param rn the RootNode
     * @return true if the file will be downloaded
     **/
    public boolean checkForDownload(RootNode rn) {

        if (GProxy.DEBUG)
            System.err.println("Called checkForDownload." + rn.getKeyword());


        /* See if this rootnode completes the query. If so, display, etc. */
        if (q_match_.addRoot(rn)) {
	    if (GProxy.DEBUG) {
	        System.err.print("QueryParser: Displaying RootNode: "); 
                System.err.print(rn.toString() + "\n");
            }

            if (displayed_roots_.get(rn.getRootINodeKey()) == null) {
                displayed_roots_.put(rn.getRootINodeKey(), 
                                     rn.getRootINodeKey());
	        qm_.display(rn);
            }
	    return true; 
        }

        /* It either doesn't complete the query or makes it false. */
        else {
            if (GProxy.DEBUG) {
                System.err.print("QueryParser: Not Displaying RootNode: "); 
                System.err.print(rn.toString() + " - Not satisfied.\n");    
            }
            return false;
        }
    }

    /**
     * Prints resulting arrays for debugging.
     **/
    public void printResults() {
        System.out.println("Whole string: ");
        for (Enumeration e = query_list_.elements(); e.hasMoreElements(); )
            System.out.println("\t " + (String)(e.nextElement()));

        System.out.println("ANDs: ");
        for (Enumeration e = and_queries_.elements(); e.hasMoreElements(); )
            System.out.println("\t " + (String)(e.nextElement()));

        System.out.println("NOTs: ");
        for (Enumeration e = not_queries_.elements(); e.hasMoreElements(); )
            System.out.println("\t " + (String)(e.nextElement()));
    }


    /**
     * Class is used by the QueryParser to keep track of
     * how arriving answers to subqueries satisfy the overall query.
     * In general, once a given root_inode is associated with all of
     * the inclusive search terms, the object's methods can indicate
     * that this file is a match. However, if at any point a root inode
     * is associated with a NOT query, it will from that point on be
     * removed from any valid query result indicators.
     **/
    class QMatches {

        /**
         * Hashtable of all of the roots encountered so far matching some
         * keyword
         **/
        private final Hashtable root_tbl_;

        /**
         * Stores values for root INodes - if the INode has satisfied
         * all AND keywords, it will be entered with a value of "true". If
         * it EVER corresponds to ANY NOT keyword, it will be entered with
         * a value of "false". Otherwise, it will not be in the table.
         **/
        private final Hashtable root_completes_;

        /**
         * The constructor. This builds the new hashtables for the
         * 2-dimensional hashtable which associates root inodes with
         * matching keywords, as well as the hashtable that indicates
         * whether or not a root_node leads to a satisfactory, invalid,
         * or incomplete match. (See root_completes_ comments above)
         **/
        public QMatches() {
            this.root_tbl_ = new Hashtable();
            this.root_completes_ = new Hashtable();
        }

        /**
         * The most important method - this adds an incoming root node with
         * its associated root inode information to the data structures.
         * It checks to see if this RootNode completes the query set, or
         * satisfies a NOT subquery and invalidates the node for future
         * consideration. This should be the only method, aside from the
         * constructor, that really needs to be called directly on this
         * object.
         * 
         * @param rn RootNode returned by the QueryModel for one of the
         *           keywords requested
         * @return true if this RootNode's root inode satisfies the query,
         *         false if the query is incomplete or invalid by means of NOT 
         **/
        public boolean addRoot(RootNode rn) {

            /* Get this root inode's keyword hashtable */
            Hashtable root_ht = 
                      (Hashtable)(root_tbl_.get(rn.getRootINodeKey()));

            /* Make a new one for it if one doesn't already exist */
            if (root_ht == null) {
                root_ht = new Hashtable();
                root_tbl_.put(rn.getRootINodeKey(), root_ht);
            }

            /* Put the keyword from this Node into the root inodes HT */
            root_ht.put(rn.getKeyword(), rn);

            /* See if this keyword completed the query */
            return checkComplete(rn);

        } 

 
        /**
         * Checks to see if this root inode has ever satisfied a NOT in
         * the past.
         * @param rn RootNode returned by the QueryModel for one of the
         *           keywords requested
         * @return true if this root inode has ever satisfied a NOT query
         *         false otherwise
         **/
        public boolean isInvalid(RootNode rn) {
            Boolean is_ok = (Boolean)root_completes_.get(rn.getRootINodeKey());
            if (is_ok == null || is_ok.booleanValue())
                return false;
            return true;
        }
        

        /**
         * Checks to see if this root inode has ever satisfied all ANDs
         * the past and never any NOTs.
         * @param rn RootNode returned by the QueryModel for one of the
         *           keywords requested
         * @return false if this has ever satisfied a NOT query or does not
         *         satisfy all AND queries, and true otherwise.
         **/
        public boolean isComplete(RootNode rn) {
            Boolean is_ok = (Boolean)root_completes_.get(rn.getRootINodeKey());
            if (is_ok == null || !(is_ok.booleanValue()))
                return false;
            return true;
        }

        /**
         * Sets the appropriate flags to indicate this root inode satisfies
         * a NOT subquery and is now invalid.
         *
         * @param rn RootNode whose root inode now satisfies a NOT query
         **/
        public void makeInvalid(RootNode rn) {
            root_completes_.put(rn.getRootINodeKey(), new Boolean(false));
        }

        /**
         * Checks to see if the root inode satisfies any NOT subqueries
         * within our query. It first checks to see if we're already
         * invalid. If we're not, we check the RootNode's keyword against
         * the NOT list. If it matches, we set the appropriate invalid
         * flags for future queries.
         * @param rn RootNode which contains the root inode whose validity
         *           we're checking
         * @return true if known or found to be invalid,
         *         false otherwise
         **/
        public boolean checkInvalid(RootNode rn) {
            if (isInvalid(rn))
                return true;

            Hashtable root_ht = (Hashtable) root_tbl_.get(rn.getRootINodeKey());

            String next_not = null;
            
            /* NOT list. Stop if we find ANY. */
            Enumeration e = not_queries_.elements(); 
	    while (e.hasMoreElements()) {            
                next_not = (String)(e.nextElement());
                /* Found one! Set completion flag to false so we know */
                /* to give up on future completion queries.           */
                if (root_ht.get(next_not) != null) {
                    makeInvalid(rn);
                    return true;
                }
            }   

            return false;
        } 
 

        /**
         * Sets the appropriate flags to indicate that, IF this root inode
         * has never yet satisfied a NOT query, this root inode now
         * satisfies all AND queries and is thus complete.
         *
         * @param RootNode whose root inode now satisfies all AND queries
         **/
        public void makeComplete(RootNode rn) {
            if (!isInvalid(rn))
                root_completes_.put(rn.getRootINodeKey(), new Boolean(true));
        }

        /**
         * Checks to see if the root inode satisfies all AND subqueries
         * within our query. It first checks to see if we're invalid. 
         * If we're not, we check to see if we're already complete. If this
         * is not the case, we check the RootNode's keyword against
         * the AND list. If it matches the whole list, we set the appropriate 
         * completion flag for future queries.
         * @param rn RootNode which contains the root inode whose completion 
         *           we're checking
         * @return false if known or found to be invalid or incomplete,
         *         true otherwise (i.e. complete and valid)
         **/
        public boolean checkComplete(RootNode rn) {
            /* if we're invalid, it doesn't matter if we're complete. */
            if (checkInvalid(rn))
                return false; 

            /* if we're already complete, don't check again. */
            if (isComplete(rn))
                return true;

            Hashtable root_ht 
		= (Hashtable)root_tbl_.get(rn.getRootINodeKey());
            String next_and = null;
            
            /* AND list. This is inefficient, but won't be TOO long */
            Enumeration e = and_queries_.elements(); 
	    while (e.hasMoreElements()) {            
                next_and = (String)(e.nextElement());

                /* At least one is missing. We're incomplete. Bail. */
                if (root_ht.get(next_and) == null) 
                    return false;                
            }

            /* We found all of them...set the completion */
            /* flag so we don't do this all over again.  */ 
            makeComplete(rn);
            return true;    
        }

    }  // end QMatches
  
} // end QueryParse
