/**
 * This represents the root node of the file tree and is used to collect all
 * of the intermediate and data nodes.
 * @file gproxy/RNode.java
 **/
package gproxy;

import java.io.*;
import java.text.*;
import java.util.Hashtable;

/**
 * This represents the root node of the file tree and
 * is used to collect all of the intermediate and data nodes.
 * <p>
 * A RNode is created for each download request.
 * @author Krista Bennett
 * @author Christian Grothoff
 **/
public class RNode 
    extends Constants 
    implements NodeInterface {

    /**
     * communication link to GNet
     **/
    private RequestManager rm_;

    /**
     * Link to the display.
     **/
    private DownloadModel dm_;
    
    /**
     * stream used to output the file to disk 
     **/
    private RandomAccessFile dos_;

    /**
     * length of file to be retrieved. Used for
     * user-feedback.
     **/
    private final long fileLength_;

    /**
     * Description of the file
     **/
    private final String description_;

    /**
     * Depth of the tree representing this file.
     **/
    private final int treeDepth_;

    /**
     * Key to obtain the file.
     **/
    private final byte[] key_;

    /**
     * The hash of the key.
     **/
    private final byte[] request_;

    /**
     * The top-Inodes.
     **/
    private Node root_;

    /**
     * Reference to the enclosing class.
     **/
    private final RNode THIS;

    /**
     * The CRC of the enclosing node.
     **/
    private final int crcRoot;

    /**
     * Total amount of data (LEAF NODES) downloaded 
     * so far for this rootnode
     **/
    private int bytes_downloaded_;

    /**
     * This represents the root node of the file
     * tree and is used to collect all
     * of the intermediate and data nodes.
     * @param descr a textual description of the node
     * @param key keyword representing this file 
     * @param fileLength the length of the file
     * @param crc the CRC of the enclosing RootNode
     * @param fileName the name of the file
     **/
    public RNode (String descr,
		  byte[] key,
		  long fileLength,
		  int crc, 
		  String fileName) {
	this.description_ = descr;
	this.fileLength_ = fileLength;
	this.key_ = key;
	this.request_ = CFuncts.hash(key_);
	this.THIS = this;
        bytes_downloaded_ = 0;
	int treeDepth = 0; 
	long fl = CONTENT_BLOCK_SIZE;
	while (fl < fileLength_) {
	    treeDepth++;
	    fl *= NODES_PER_INODE;
	}
	this.treeDepth_ = treeDepth;
	this.crcRoot = crc;
	// System.out.println("Root-node crc is " + printCRC(crc));
	RandomAccessFile raf = null;
	try {
	    raf = new RandomAccessFile(new File
				       (fileName),
				       "rw"); 
	} catch (IOException e) {
	    System.err.println("ERROR: "+e);
	    System.exit(-1);
	}
	this.dos_ = raf;
	if (treeDepth_ == 0)
	    root_ = new Node(key, 0, 0, this);
	else
	    root_ = new INode(key, 0, 0, this);
    }
    
    /**
     * This method starts the download 
     * process for this file. This download will terminate
     * either when the file is downloaded completely or
     * when the abort() method is called. The method
     * will return nearly immediately, though.
     * @param rm whom to ask for data
     * @param dm whom to tell what's going on
     **/
    public void download(RequestManager rm,
			 DownloadModel dm) {
	this.rm_ = rm;
	this.dm_ = dm;
	root_.request();
    }

    /**
     * This method receives data corresponding
     * to the indicated filename (hashcode)
     * @param hash hashcode representing filename
     * @param data data corresponding to this node
     **/ 
    public synchronized void receive(byte[] hash, byte[] data) {
	throw new Error("RNode receive called!");
    }

    /**
     * Abort this download.
     **/
    public synchronized void abort() {
	root_.abort();
    }

    /**
     * Return a textual description of the node.
     **/
    public String toString() {
	return new String(description_ + "          " + 
			  formatCurrentPercent() + "%");
    }

    /**
     * How many bytes did we receive so far?
     * @return the number of bytes received
     **/
    public long received() {
	return root_.received();
    }
    
    /**
     * How many bytes did we receive so far?
     * @return the number of bytes received
     **/
    public long missing() {
	return root_.missing();
    }

    /**
     * Perform the request for the root-node.
     **/
    public void request() {
	root_.request();
    }

    /**
     * Notifiation: another child-node is complete
     **/
    public synchronized void notifyReceivedData() {
	if (root_.missing() == 0) {
	    // Yes, subtree is complete. CRC check!
	    byte[] crcBlock = new byte[4];
	    byte[] signData = root_.getSignificantData();
	    int crc = CFuncts.crc32(signData);
	    if (crcRoot != crc) {
		System.err.println
		    ("CRC (Root) error. Collision in one of the blocks!\n"+
		     "You hit the jackpot (51:2^160 chances). Please report this!\n"+
		     "You may also want to check file integrity." + 
		     printCRC(crc) + " != " + printCRC(crcRoot));
		// System.exit(-1);
	    }	    
	    root_ = null;	    
	}
	dm_.update(THIS);	
	if (root_ == null) // either we never had children, or they are all done.
	    dm_.notifyComplete(THIS); // in either case, we are finished
	
    }

    /**
     * Get percent of leaf node data already downloaded.
     * @return float representation of current percent of content downloaded
     **/
    public float getPercentDownloaded() {
       return (((float)bytes_downloaded_)/fileLength_)*100;
    }

    /**
     * Format current percent and return a string
     * @return formatted string (2 decimal places) 
     *         representing percent downloaded
     **/
    public String formatCurrentPercent() {
        DecimalFormat percentFormat 
	    = new DecimalFormat("###.##");
        return percentFormat.format(getPercentDownloaded());
    }

    /**
     * Get the length of the file that is covered by each child
     * @return the number of bytes each child of this node may
     *         cover (max, disregarding padding)
     **/
    private long getChildLen(int depth) {
	if (depth > treeDepth_)
	    return 0;
	long childlen = CONTENT_BLOCK_SIZE; 
	// how big is each child-node (full tree below)	
	for (int i=depth;i<treeDepth_-1;i++)
	    childlen *= NODES_PER_INODE; 
	    // is child-node an inner node?
	return childlen;
    }
    
    /**
     * Compute the number of children of this node,
     * acknowleding padded inodes
     * at the end of the file.
     * @return the number of children this node will 
     * have, will be less or equal to NODES_PER_INODE.
     **/
    private int getNumberOfChildren(long pos,
				    int depth) {
	if (depth > treeDepth_) 
	    return 0;
	// compute number and position of subtrees
	long childlen = getChildLen(depth);
	long npi = (((int)fileLength_) - pos + childlen - 1) 
	    / childlen; // how many child nodes?
	if (npi > NODES_PER_INODE)
	    npi = NODES_PER_INODE; // cap at maximum	   
	return (int) npi;
    }
	
    /**
     * Check that the block "data" actually
     * matches the block we are looking for (hash-wise)
     * @param len the length of the part of delta that
     *        we are expecting to be the valid part.
     * @return true if the data we have received
     *         matches our query
     **/
    private boolean hashCheck(byte[] data,
			      byte[] hash,
			      int len) {	
	byte[] dat = new byte[len];
	System.arraycopy(data,0,
			 dat,0,
			 len);
	byte[] hashCheck = CFuncts.hash(CFuncts.hash(dat));
	return CFuncts.hashEquals(hash, hashCheck);
    }

	
    /* *************** inner classes to build the tree******** */
    
    /**
     * This class represents the nodes of the file-tree,
     * either inner nodes or leaf nodes. For the root-node,
     * the subclass RNode is used. 
     * @author Christian Grothoff
     **/
    private class Node 
	implements NodeInterface {
	
	/**
	 * Parent-node in the tree.
	 **/
	final NodeInterface parent;
		
	/**
	 * Position in the tree (corresponding to the lowest
	 * position index of its leftmost child)
	 **/
	final long pos;
	
	/**
	 * This node's depth in the tree
	 **/
	final int depth;
		
	/**
	 * The key of the node (to decrypt)
	 **/
	final byte[] key;
	
	/**
	 * What is the request that maches this node?
	 **/
	final byte[] request;
	
	/**
	 * The 1024 bytes of data associated with this node
	 **/
	byte[] data;
	
	/**
	 * This creates a node.
	 * @param key key used to decrypt this node
	 * @param pos position of this node in 
	 *            this level of the tree
	 * @param depth depth of this node in the tree
	 * @param parent parent node of this node 
	 *        (used to notify parent of completed download)
	 **/
	private Node(byte[] key, long pos, int depth,
		     NodeInterface parent) {
	    this.pos     = pos; 
	    this.depth   = depth;
	    this.key     = key;
	    this.request = CFuncts.hash(key); // cached for speed
	    this.parent  = parent;
	}

	public String toString() {
	    return "Leaf(" + pos + ")";
	}

	/**
	 * This method is used to notify the node that
	 * some relevant data may have arrived.
	 * @param hash filename of data that has arrived
	 * @param data data corresponding to the hash code
	 **/
	public synchronized void receive(byte[] hash, 
					 byte[] dat) { 
	    if (!CFuncts.hashEquals(hash, request))
		throw new Error("receive called, but hash does not match!?");	    
	    this.data = CFuncts.decrypt(key, dat);
	    if (this.data == null) {
		request(); // re-issue request!
		return; // invalid data received, ignore!
	    }	    
	    if (!hashCheck(data,hash,getBlockLength())) {				
		System.err.println
		    ("Invalid packet for Leaf-Node at position " +
		     this.pos + 
		     " (hash does not decode). Ignored.");
		this.data = null;
		request(); // re-issue request
		return;
	    }
	    
	    try {
		synchronized (dos_) {
		    dos_.seek(pos); 
		    dos_.write(data,0,getBlockLength()); // write to file
                    THIS.bytes_downloaded_ += getBlockLength();
		}		
	    } catch (IOException io) {		    
		System.err.println("Error writing data to disk: "+ 
				   io);
		io.printStackTrace();
		System.exit(-1);
	    }
	    notifyReceivedData(); 
	}
    
	/**
	 * This code is called both to notify this node 
	 * that one of its child-nodes received data.
	 **/
	public void notifyReceivedData() {
	    parent.notifyReceivedData(); 
	}

	/**
	 * How many bytes did we receive so far?
	 * @return the number of bytes received
	 **/
	public long received() {
	    if (this.data != null)
		return getBlockLength();
	    else
		return 0;
	}

	/**
	 * How many bytes did we receive so far?
	 * @return the number of bytes received
	 **/
	public long missing() {
	    if (this.data == null)
		return getBlockLength();
	    else
		return 0;
	}

	/**
	 * Return the length of the block without padding
	 * @return CONTENT_BLOCK_SIZE except for
	 *         the last block if the last block is padded.	 
	 **/
	protected int getBlockLength() {
	    long delta = fileLength_ - pos;
	    if (delta > CONTENT_BLOCK_SIZE)
		delta = CONTENT_BLOCK_SIZE;
	    return (int) delta;
	}
	
	/**
	 * Get the significant part of the data (without padding)
	 * @return the data of the block, eventually getting rid
	 *         of padding first (only for the last block)
	 **/
	private final byte[] getSignificantData() {
	    int bl = getBlockLength();
	    if (bl == this.data.length)
		return this.data;
	    byte[] dat = new byte[bl];
		System.arraycopy(this.data,0,
				 dat,0,
				 bl);
		return dat;
	}
	
	/**
	 * Send out request for this node.
	 * @param priority the priority for this request (max)
	 **/
	public synchronized void request() {
	    if (data != null)
		return; // or throw error...
	    rm_.request(this,
			this.request);
	}       

	/**
	 * Abort this download.
	 **/
	public synchronized void abort() {
	    if (data == null)
		rm_.abort(this, this.request);
	}
	
     } // end of Node

    /**
     * Representation of an inner node
     **/
    private final class INode extends Node {
	
	/**
	 * Child-nodes (may have null-entries).
	 **/
	private Node[] childs_;
	
	/**
	 * Checksum of the collected children of this node
	 * (what we'd like to get)
	 **/
	private int crc32;

	/**
	 * This creates a node.
	 * @param key key used to decrypt this node
	 * @param pos position of this node in this 
	 *        level of the tree
	 * @param depth depth of this node in the tree
	 * @param parent parent node of this node (used 
	 *        to notify parent of completed download)
	 **/
	private INode(byte[] key, long pos, int depth,
		      NodeInterface parent) {
	    super(key,pos,depth,parent);
	}

	public String toString() {
	    return "Inner(" + pos + "," + depth + ")";
	}

	/**
	 * This method is used to notify the node that
	 * some relevant data may have arrived.
	 * @param hash filename of data that has arrived
	 * @param dat data corresponding to the hash code
	 **/
	public synchronized void receive(byte[] hash, 
					 byte[] dat) { 
	    if (!CFuncts.hashEquals(hash, request))
		throw new Error
		    ("receive called, but hash does not match!?");
	    // System.out.println("received data for " + id);
	    // we can now resolve this node!
	    this.data = CFuncts.decrypt(key,dat);
	    if (data == null)
		return; // invalid packet, did not decrypt
	    
	    if (!hashCheck(data,hash,getBlockLength())) { 
		System.err.println
		    ("Invalid packet for Inode (hash does not decode). Ignored.");
		this.data = null;
		this.request(); // re-issue request!
		return;		
	    }	    
	    
	    long childlen = getChildLen(depth);
	    childs_ = new Node[getNumberOfChildren(pos,depth)]; 
	    for (int i=0;i<childs_.length;i++) {
		byte[] key = new byte[HASH_SIZE];
		System.arraycopy(data, HASH_SIZE*i,
				 key, 0,
				 HASH_SIZE);
		if (depth+1==treeDepth_)
		    childs_[i] = new Node(key, pos + childlen*i,
					  depth + 1, this);
		else
		    childs_[i] = new INode(key, pos + childlen*i,
					   depth+1,this);
		childs_[i].request();
	    }	    
	    crc32 = 
		Converter.convertEndian
		(Converter.getInt(data,childs_.length*HASH_SIZE));
	    //System.out.println("expecting crc " + printCRC(crc32) +
	    //		       " as found at " + childs_.length*HASH_SIZE +
	    //		       " for node " + this.toString());
	}

	/**
	 * Abort this download.
	 **/
	public synchronized void abort() {
	    if (childs_ != null) 
		for (int i=0;i<childs_.length;i++)
		    if (childs_[i] != null)
			childs_[i].abort();
	    if (data == null)
		rm_.abort(this, request_);
	}
	
	/**
	 * How many bytes did we receive so far?
	 * @return the number of bytes received
	 **/
	public long received() {
	    long received = CONTENT_BLOCK_SIZE;
	    if (childs_ != null) 
		for (int i=0;i<childs_.length;i++)
		    if (childs_[i] != null)
			received += childs_[i].received();
	    return received;
	}

	/**
	 * How may bytes are missing for this node to
	 * be complete?
	 **/
	public long missing() {
	    if (this.data == null) {
		// this is broken: it does
		// not contain the size of IBlocks.
		// FIXME!
		return 
		    getChildLen(depth)*getNumberOfChildren(pos,depth) 
		    + CONTENT_BLOCK_SIZE;
	    } else {
		long missing = 0;
		if (childs_ != null) 
		    for (int i=0;i<childs_.length;i++)
			if (childs_[i] != null)
			    missing += childs_[i].missing();
		return missing;
	    }
	}

 	/**
	 * Return the length of the block without padding that
	 * matches actual content of the INode.
	 * @return the length in bytes, including
	 *         the hashcode and the indirection-hashes
	 **/
	protected int getBlockLength() {
	    int npi = getNumberOfChildren(pos,depth);
	    return npi*HASH_SIZE+4;
	}

        /**
         * This code is called both to notify 
	 * this node that another of its children is complete
         * as well as to notify this node's parent that
         * this node is complete when all of its
         * children's downloads are complete.
         **/
        public synchronized void notifyReceivedData() {
	    if (childs_ == null) 
		return; // very rare, but may occur due to concurrency
	    int childDone = 0;
	    for (int i=0;i<childs_.length;i++)
		if (childs_[i].missing() == 0)
		    childDone++;
            if (childDone == childs_.length) {
		// Yes, subtree is complete. CRC check!
		byte[] crcBlock = new byte[childs_.length * 4];
		int crc = 0; // start value
		for (int i=0;i<childs_.length;i++) {
		    byte[] sigData 
			= childs_[i].getSignificantData();		   
		    int blockCrc = Converter.reverseEndian
			(CFuncts.crc32
			 (sigData));
		    Converter.setInt(crcBlock,i*4, blockCrc);
		}
		crc = CFuncts.crc32(crcBlock);
		if (crc32 != crc) {
		    System.err.println
			("CRC error. Collision in one of the blocks!\n"+
			 "You hit the jackpot (51:2^160 chances). Please report if you can reproduce this.\n" +
			 "You may also want to check the integrity of your file...\n ." + 
			 printCRC(crc) + 
			 " != " + 
			 printCRC(crc32));
		    
		    System.err.println("INode starts at " + pos + 
		    		       ", depth is " + depth);
		    for (int i=0;i<childs_.length;i++)
			System.err.println("crc of childs_[" + i + 
					   "].significant = " +
					   printCRC
					   (CFuncts.crc32
					    (childs_[i].getSignificantData())) +
					   " crc over " 
					   + childs_[i].getSignificantData().length);
		    /* new Error("test").printStackTrace();
		       System.exit(-1); */
		}			
		childs_ = null; // GC!
	    }
            parent.notifyReceivedData();
        }       
    } // end of INode

    static String printCRC(int crc) {
	return (crc&255) + "/" + (Converter.reverseEndian(crc)&255);
    }
 
} // end of RootNode
