/**
 * @file gproxy/RequestManager.java
 **/
package gproxy;

import java.util.Vector;

/**
 * The RequestManager keeps track and re-issues
 * requests for Content-Nodes.
 * @author Christian Grothoff
 **/
public class RequestManager
    extends Thread 
    implements Listener {

    /**
     * How many requests do we have pending (max)
     * Must be > 50 to make sure we trigger flushing buffers
     * at the node.
     **/
    private final static int MAX_REQUESTS = 64;

    /**
     * The BuildThread - where we send our requests.
     **/
    private final BuildThread bt_;

    /**
     * Requests we know of.
     **/
    private final Vector queue_;

    /**
     * Recently send requests.
     **/
    private final Request[] send_;

    /**
     * Current "good" TTL (initial) [64s]
     **/
    private int startTTL = 64;

    /**
     * Create a new RequestManager. Starts the
     * RequestManager thread.
     **/
    public RequestManager(BuildThread bt) {
	this.bt_ = bt;
	this.bt_.addListener(this);
	this.send_ = new Request[MAX_REQUESTS];
	this.queue_ = new Vector(MAX_REQUESTS);
	start();
    }

    /**
     * Main loop of the RequestManager.
     **/
    public void run() {
	while (true) {
	    for (int i=0;i<send_.length;i++) {
		// keep requesting
		synchronized(send_) {		
		    while (i<send_.length) {
			if (send_[i] == null) {
			    synchronized(queue_) {
				if (!queue_.isEmpty()) 
				    send_[i] = (Request) 
					queue_.remove(0); 
			    }			
			}		
			if ((send_[i] != null) &&
			    (send_[i].lastTime_ < System.currentTimeMillis() - 1000*startTTL)) {
			    send_[i].lastTime_ = System.currentTimeMillis();
			    send_[i].request();
			    i++;
			    break; // stop after first request
			}
			i++;
		    } // end while i<send_.length
		} // end sync send_
	    } // for i
	    CFuncts.sleep(2);  // 2 secs
	} // while(true)
    }

    /**
     * Add a request.
     * @param l which object should be notified on completion?
     * @param query what is the query?
     **/
    public void request(Listener l,
			byte[] query) {
	synchronized(send_) {		
	    for (int i=0;i<send_.length;i++) 
		if (send_[i] == null) {
		    send_[i] = new Request(l,query);
		    send_[i].lastTime_ = System.currentTimeMillis();
		    send_[i].request();
		    return;
		}	    
	} // sync send_
	synchronized (queue_) {
	    queue_.add(new Request(l,query));
	} 
    }

    /**
     * Abort a request
     * @param l which object did the original request
     * @param query what was the query
     **/
    public void abort(Listener l,
		      byte[] query) {
	Request r = new Request(l, query);
	synchronized(queue_) {
	    queue_.remove(r);
	}
	synchronized(send_) {
	    for (int i=0;i<send_.length;i++)
		if (r.equals(send_[i]))
		    send_[i] = null;
	}
    }
  
    /**
     * This method receives data corresponding to the indicated filename (hashcode)
     * Find the Listener that scheduled this request and
     * drop it from the list of pending requests.
     * @param hash hashcode representing filename
     * @param data data corresponding to this node
     **/ 
    public void receive(byte[] hash, byte[] data) {
	synchronized(send_) {
	    for (int i=0;i<send_.length;i++) 
		if (send_[i] != null)
		    if (CFuncts.hashEquals(hash, send_[i].query_)) {
			startTTL = (int) ((System.currentTimeMillis()-send_[i].lastTime_)/1000+4); // new TTL = time it took 
			send_[i].l_.receive(hash,data);
			send_[i] = null;
			synchronized(queue_) {
			    if (!queue_.isEmpty()) {
				send_[i] = (Request) queue_.remove(0);			    
				send_[i].lastTime_ = System.currentTimeMillis();
				send_[i].request(); // request now - no wait!
			    } else {
				// few queries pending, re-first
				int k=0;
				while ((k < send_.length) &&
				       (send_[k] != null))
				    k++;
				if (send_[k] != null) {
				    send_[i].lastTime_ = System.currentTimeMillis();
				    send_[k].request();
				}
			    } // else
			} // sync queue
		    }  // if hashEquals    
	} // sync send_
    }

    /**
     * Class representing pending requests.
     * @author Christian Grothoff
     **/
    private class Request {
	/**
	 * Who made this request?
	 **/
	private final Listener l_;

	/**
	 * What is the question?
	 **/
	private final byte[] query_;

	/**
	 * Current priority
	 **/
	private int prio_;

	/**
	 * Current time-to-live
	 **/
	private int ttl_;
	
	/**
	 * When did we issue this query the last time?
	 **/
	private long lastTime_;
	
	/**
	 * Create a request object.
	 * @param l who asks
	 * @param query what is the question
	 **/
	private Request(Listener l,
		byte[] query) {
	    this.l_     = l;
	    this.query_ = query;
	    this.ttl_   = startTTL;
	    this.prio_  = 0;
	}
	/**
	 * Issue the request.
	 **/
	private void request() {
	    bt_.sendData(query_, prio_, ttl_);
	    prio_++;
	    ttl_+=10;
	}
	
	/**
	 * Are these two requests identical?
	 **/
	public boolean equals(Object o) {
	    if (o == null)
		return false;
	    Request oo = (Request)o;
	    return ((oo.l_ == l_) &&
		(CFuncts.hashEquals(oo.query_,query_)));
	}

    } // end of Request

} // end of RequestManager
