/**
 * RootNode represents the RootNode of a file,
 * with description, hash and file-length 
 * information.
 * @file gproxy/RootNode.java
 **/
package gproxy;

/**
 * RootNode represents the RootNode of a file,
 * with description, hash and file-length 
 * information.
 * @author Christian Grothoff
 * @author Krista Bennett
 **/
public class RootNode extends Constants {

    /**
     * The hash that was send with this
     * root-node.
     **/
    private final byte[] hash_;

    /**
     * The data of this root-node.
     **/
    private byte[] data_;

    /**
     * The keyword matching this
     * root-node. (null if no keyword is known).
     **/
    private String keyword_;

    /**
     * Description of the root-node.
     * Null if not yet decoded.
     **/
    private String description_;

    /**
     * Key of the INode at the top of the
     * file.
     **/
    private byte[] key_;

    /**
     * Length of the correspoinding field
     **/
    private long fileLength_;

    /**
     * CRC checksum of the root-node
     **/
    private int crc32_;

    /**
     * Create a new RootNode.
     * @param hash the hash that came with the node
     * @param data the data of the node.
     **/
    public RootNode(byte[] hash,
		    byte[] data) {
	this.hash_ = hash;
	this.data_ = data;
    }

    /**
     * Try decrypting this node with the give
     * keyword.
     * @param keyword the keyword
     * @return true if the decryption succeeded.
     **/
    public synchronized boolean decrypt(String keyword) {
	if (keyword_ != null)
	    throw new Error("RootNode already decrypted!");	
	byte[] h2 = CFuncts.hash(CFuncts.hash(keyword.getBytes()));
	if (!CFuncts.hashEquals(h2, hash_))
	    return false;
	keyword_ = keyword;
	data_ = CFuncts.decrypt(CFuncts.hash(keyword.getBytes()),
				data_);
	int descLen 
	    = Converter.convertEndian
	    (Converter.getChar(data_,ROOT_DESCLEN_OFFSET));	     
	if (descLen > data_.length - ROOT_DESC_OFFSET) 
	    throw new Error("Description too long! " + descLen);
	
	this.description_ 
	    = new String(data_,ROOT_DESC_OFFSET,descLen); 
	if (this.description_ == null)
	    throw new Error("Description could not be created: " + descLen);
	key_ = new byte[HASH_SIZE];
	System.arraycopy(data_,ROOT_HASH_OFFSET,
			 key_,0,
			 HASH_SIZE);
	fileLength_ =
	    Converter.reverseEndian
	    (Converter.getInt(data_,FILE_LENGTH_OFFSET));                 
	if (fileLength_ <= 0)
	    throw new Error("Bad filelength: " + fileLength_);
	crc32_ = Converter.convertEndian
	    (Converter.getInt(data_,
			      ROOT_CRC32_OFFSET));

	return true; // successful decryption!
    }

    /**
     * Return a description of this file.
     **/
    public String toString() {
	StringBuffer res = new StringBuffer(description_);
	while (res.length() < 40)
	    res.append(" ");
	res.append(" Size: ");
	String siz = Long.toString(fileLength_);
	while (siz.length() + res.length() < 52)
	    res.append(" ");
	res.append(siz);
	return res.toString();
    }
    
    /**
     * Create the RNode that can download this
     * file.
     **/
    public RNode getDownloadNode(String fileName) {
	return new RNode(description_, 
			 key_, fileLength_,
			 crc32_, fileName);
    }        

    /**
     * Returns keyword for this file
     * @return keyword String associated with this RootNode
     **/
    public String getKeyword() {
        return new String(keyword_);
    }

    /**
     * Returns key of main INode for this file
     * @return hex String conversion of the root inode name
     **/
    public String getRootINodeKey() {
        return CFuncts.bytes2hexString(key_);
    }


} // end of RootNode
