/**
 * insertFile - breaks file meant for
 *              insertion into blocks, prepares
 *              for insertion into network
 * @author Krista Bennett
 * @author Christian Grothoff
 * @file gproxy/insertfile.c
 **/

#include "insertfile.h"
#include "sessionkey.h"
#include "lookup.h"

/* #define DEBUG 1 */
/* #define DEBUG_CRC 1 */

/**
 * Print the current indirection table.
 **/
void printTable(HashCode160 * hashBuffer[], 
		unsigned short * pos,
		int treeDepth) {
  int i;
  int j;
  HexName hex;
  
  print("INODE TABLE:");
  for (i=0;i<= treeDepth;i++) {
    print("\nTable %d: ",i);
    for (j=0;j<pos[i];j+=sizeof(HashCode160)) {
      hash2hex(&hashBuffer[i][j],&hex);
      print("%d: %s ",j, &hex);
    }
  }
  print("\nEND TABLE\n");

}
   
/**
 * Adds random information at the end of a block portion
 *
 * @param data represents the data block
 * @param actual_length represents the length of 
 *        the useful information in the
 *	  block; the rest of the bytes should be replaced by
 *	  random information
 **/
void padCONTENT_SIZE(CONTENT_Block * data, int actual_length) {
  while (actual_length < CONTENT_SIZE)
    data->content[actual_length++] = (char) rand();
}

/**
 * Map the given file into memory.
 * @param flen pointer to a location where to
 *        store the length of the file
 * @return a pointer to the mmapped file.
 **/
void * openFile(size_t * flen, 
		FileName filename) {
  int fd;
  struct stat st; /* file info */
  void * res;

  /* open file, return error if unable to open for reading */
  if ((fd = open(filename,O_RDONLY)) < 0) 
    errexit("Insertion error: unable to open %s\n", filename);
  stat(filename, &st); /* get file size */   
  *flen = st.st_size;
  if (*flen <= 0) 
    return NULL;
  res = mmap(NULL, /* no preferred address */
	     *flen, /* file length */
	     PROT_READ, /* we will only read */
	     MAP_SHARED, /* others may access this ro area, too */
	     fd, 
	     0 /* offset */);
  close(fd);
  return res;
}

/**
 * Encrypt a block of data using its hashcode.
 * @param block 1024-bit data to be encrypted
 * @param hc hashcode of the data we're going to encrypt
 * @return 1024-bit block of encrypted data
 **/
void encodeBlockWithHash(CONTENT_Block * block,
			 HashCode160 * hc,
			 CONTENT_Block * result) {
#ifdef DEBUG
  print("Encrypting %s with %s\n", 
	 printMem(block,CONTENT_SIZE),
	 printHASH(hc));
#endif
  encryptContent(block, hc, result); 
}

#define NODES_PER_INODE ((int)(CONTENT_SIZE/sizeof(HashCode160)))

/**
 * Compute the depth of the tree.
 * @param flen file length from which to compute the depth
 * @return depth of the tree
 **/
unsigned char computeDepth(size_t flen) {
  unsigned char treeDepth;
  long fl;

  treeDepth = 0;
  fl = CONTENT_SIZE;
  while (fl < flen) {
    treeDepth++;
    fl = fl * NODES_PER_INODE;
  }  
  return treeDepth;
}

/**
 * Write a block to the disk. Use H(H(block)(
 * for the filename. Encrypt with H(block). 
 * If the block is less than CONTENT_SIZE bytes, pad the last
 * CONTENT_SIZE-len bytes. Do not include the padded
 * part in the hash.
 * @param block the block to write
 * @param len the part of the block that is valid (not to pad)
 * @return single hash of the data
 **/
void writeBlock(CONTENT_Block * block, 
		int len,
		HashCode160 * hc) {
  HashCode160 hcSquare;
  HexName hex;
  CONTENT_Block eBlock;

  hash(block, len, hc);
  hash(hc, sizeof(HashCode160), &hcSquare);
  hash2hex(&hcSquare, &hex);
  encodeBlockWithHash(block, hc, &eBlock);
#ifdef DEBUG
  print("Writing file %s for hash %s\n",hex,hash2hex(hc));
#endif
  writeContent(&hex, &eBlock);
}

/**
 * Write an indirection block with the given crc
 * under its hashcode to the content directory.
 * @param block the data in the indirection block
 * @param pos the length of the data that is used
 * @param crc32 the CRC to store with the block 
 * @param single hash of the block written
 **/
void writeIndirection(CONTENT_Block * block,
		      unsigned short pos,
		      Ulong crc32,
		      HashCode160 * hc) {
  ((Ulong *)block)[pos/sizeof(Ulong)] = crc32;
#ifdef DEBUG_CRC
  print("Expecting crc %d\n",crc32&255);
#endif
  writeBlock(block,pos+sizeof(Ulong),hc);
}

/**
 * Walk over the indirection nodes (inodes)
 * up from a leaf to the top, adding hashcodes
 * to the node above if the node below is full.
 * Also computes the CRC values.
 * @param hashBuffer list of tree-depth hash buffers 
 * @param crc32Buffer list of tree-depth CRC buffers
 * @param pos list of tree-depth write-positions in the 
 *        hash buffer (and implicitly in the crc buffer)
 * @param crc the CRC code that should be appended at level j
 * @param hc the hash code that should be appended at level j
 * @param j the current level of traversal
 **/
void walkIndirections(CONTENT_Block hashBuffer[],
		      Ulong ** crc32Buffer,
		      unsigned short * pos,
		      Ulong crc,
		      HashCode160 * hc,
		      unsigned char j,
		      int offset) {
#ifdef DEBUG
  int i;
#endif
#ifdef DEBUG_CRC
  int cc;
#endif

  /* append hash and CRC, increase pos accordingly */
  memcpy(&hashBuffer[j].content[pos[j]],
	 hc,sizeof(HashCode160));
  crc32Buffer[j][pos[j]/sizeof(HashCode160)]=crc;
  pos[j] += sizeof(HashCode160);

  /* while buffer of level j is full, flush buffer to file and
     propagate summary up to level j+1. */
  while (pos[j]+sizeof(HashCode160) > CONTENT_SIZE) {
#ifdef DEBUG
    print("walkingIndirections level %d hash %s\n",j,
	   printHASH(*hc));
#endif

    /* compute CRC of the individual CRCs (51 CRCs to 1 CRC) */
    crc = crc32N(crc32Buffer[j], 
		 sizeof(Ulong) *
		 (CONTENT_SIZE / sizeof(HashCode160)));
#ifdef DEBUG_CRC
    //    if (j>0) {
    print("Level %d: at offset %d CRC of ",offset,j);
    for (cc=0;cc<(CONTENT_SIZE / sizeof(HashCode160));cc++)
      print("%d ",crc32Buffer[j][cc] &255);
    print(" is %d.\n",crc & 255);
    //}
#endif    

    writeIndirection(&hashBuffer[j], pos[j],
		     crc, hc); /* flush to file */
#ifdef DEBUG
    print("Hash at %d level %d is %s\n", pos[j], j,
	   printHASH(*hc));      
    for (i=0;i<51;i++) 
      print("INode entry %d: %s\n",i,
	     printHASH(&hashBuffer[j][i*20]));
#endif
   
    pos[j] = 0;
    crc = crc32N(&hashBuffer[j], CONTENT_SIZE);
    j++;
#ifdef DEBUG
    print("Writing hash of INode to hashBuffer[%d][%d]: %s\n",
	   j,pos[j],
	   printHASH(hc));
#endif
    memcpy(&hashBuffer[j].content[pos[j]],
	   hc,sizeof(HashCode160));
    crc32Buffer[j][pos[j]/sizeof(HashCode160)]=crc;
    pos[j] += sizeof(HashCode160);    
  } 
}


/**
 * Creates root node for the tree and writes the top-level tree node.
 * Root looks like this (when written to file - first 32 bits that
 * are NOT written to file contain content block length for hash and
 * writeFile functions):
 *
 *    32     32     160      strlen of desc     CONTENT_SIZE - the rest
 * ____________________________________________________________
 * |      |      | Hash  |                     | 
 * | File | Root | of    | Content description |  Padding    
 * |      | Node | Root  |                     |
 * | Size | CRC  | Indir |                     |  (if needed)
 * |      |      | Node  |                     |
 * ------------------------------------------------------------
 *
 * @param hc the hashcode of the top-node of the node-tree
 * @param keywords the keywords for the root-node
 * @param num_keys number of keywords to associated with the root node
 * @param description the description of the content
 * @param flen the length of the file
 * @param crc the CRC checksum for the file
 **/
void writeRootNode(HashCode160 * hc,
		   char ** keywords,
                   int num_keys,
		   unsigned char * description,
		   size_t flen,
		   Ulong crc) {
  RootNode rc;
  BLOCK_LENGTH dlen;
  HashCode160 hcSquare;
  HexName hex;
  CONTENT_Block block;
  int i;

#ifdef DEBUG
  print("Writing root-node with hash %s\n",hash2hex(hc));
#endif
  dlen = strlen(description);
  /* truncate desc if necessary */
  if (dlen > MAX_DESC_LEN)
     dlen = MAX_DESC_LEN;
  /* create root node (this part is keyword independent) */
  rc.file_length = flen;
  rc.crc = crc;
  memcpy(&rc.hashCode, hc, sizeof(HashCode160));
  rc.descSize = dlen;
  memcpy(&rc.description[0], description, dlen);
  padCONTENT_SIZE((CONTENT_Block*)&rc,
		  sizeof(RootNode)-MAX_DESC_LEN+dlen);
 
  /* for each keyword, create a new root node and write the file */
  for (i = 0; i < num_keys; i++) {
    hash(keywords[i],strlen(keywords[i]),hc);
    hash(hc,sizeof(HashCode160),&hcSquare);  
    hash2hex(&hcSquare, &hex);
    /* encrypt rc with "hc" here */
    encodeBlockWithHash((CONTENT_Block*)&rc, hc, &block); 
#ifdef DEBUG
    print("writing root node (%s)\n",&hex);
#endif
    writeContent(&hex, &block);
  }
}

/**
 * De-facto main method. Inserts a file under the given
 * name into the local gnet node.\
 * @param filename the name of the (incomming/source) file
 * @param description the description of the file
 * @param keywords the keywords that shall be used to retrieve the file
 * @param num_keys the number of keywords to be associated with the file
 * @param encode_content: != 0, if we are going to encrypt, encode, and write      
 *                        the content portions themselves
 *                        0, if we are simply going to create the indirection 
 *                        nodes for the file (for on-demand encoding later)
 * @return 0: ok, -1: file not found
 **/
int insertFile(FileName filename, 
	       unsigned char * description,
               char ** keywords,
	       int num_keys,
               unsigned int encode_content) {

  unsigned char i;
  unsigned char treeDepth;
  size_t flen;
  size_t fpos;
  unsigned char * mmappedFile;
  CONTENT_Block * hashBuffer;
  Ulong ** crc32Buffer;
  unsigned short * pos;
  HashCode160 hc;
  CONTENT_Block * block = NULL;   
  Ulong crc = 0;
  ContentIndex entry;
#ifdef DEBUG_CRC
  int cc;
#endif

  mmappedFile = openFile(&flen, filename);
  if (mmappedFile == NULL)
    return -1; /* failed */

  if (flen == 0)
    return -1; /* file empty - failed */
  treeDepth = computeDepth(flen);

  /* allocate memory */
  hashBuffer 
    = malloc(sizeof(CONTENT_Block) * (treeDepth+1));
  crc32Buffer 
    = malloc(sizeof(Ulong*) * (treeDepth+1));
  pos 
    = malloc(sizeof(unsigned short) * (treeDepth+1));

  /* initialize buffers and offsets */
  for (i=0;i<treeDepth+1;i++) {
    crc32Buffer[i]
      = (Ulong*) malloc(sizeof(Ulong) *
			(CONTENT_SIZE / sizeof(HashCode160)));
   pos[i] = 0;
  }

  /* walk through nodes - start at leaf, walk up complete nodes */
  fpos = 0;

  while (fpos < flen) {
    if (flen - fpos >= CONTENT_SIZE) {
      block = (CONTENT_Block*) &mmappedFile[fpos];    
      crc = crc32N(block,CONTENT_SIZE);
      /* write to file? or just create indirection nodes? */
      if (encode_content == 0) 
        writeBlock(block,CONTENT_SIZE, &hc);
      else {
        hash(block, CONTENT_SIZE, &hc);	  
	/* add entry to database! */
	hash(&hc,
	     sizeof(HashCode160),
	     &entry.doubleHash);
	entry.importance = 0xFFFF;
	entry.fileNameIndex = encode_content;
	entry.fileOffset = fpos;
	addEntry(&entry);
      }
#ifdef DEBUG
      print("Hash at %d is %s\n", fpos,
	    printHASH(&hc));
#endif
    } 
    /* last (incomplete) block */
    else { 					
      block = malloc(CONTENT_SIZE);
      padCONTENT_SIZE(block,flen-fpos);
      memcpy(block, &mmappedFile[fpos], flen-fpos);
      crc = crc32N(block,flen-fpos);
      writeBlock(block,flen-fpos,&hc);
#ifdef DEBUG
      print("Hash at %d is %s\n", fpos,
	    printHASH(&hc));
#endif
    }
    if (treeDepth > 0)
      walkIndirections(hashBuffer, crc32Buffer, 
		       pos, crc, &hc, 0, fpos);   
#ifdef DEBUG
    printTable(hashBuffer, pos, treeDepth);
#endif
    fpos += CONTENT_SIZE;
  }
  /* free leaf-specific stuff */
  if ((flen & (CONTENT_SIZE-1)) > 0)
    /* CONTENT_SIZE MUST be a power of 2! */
    free(block); /* free padded block if one was created in the last iteration */
  munmap(mmappedFile,flen);

  /* complete tree... */
  for (i=0;i<treeDepth;i++) {
    if (pos[i] == 0)
      continue; /* nothing to complete: full node! */
    crc = crc32N(crc32Buffer[i], 
		 sizeof(Ulong) *
		 (pos[i] / sizeof(HashCode160)));
    padCONTENT_SIZE(&hashBuffer[i], pos[i]);
    writeIndirection(&hashBuffer[i], pos[i], 
		     crc, &hc);
    crc = crc32N(&hashBuffer[i], pos[i]+sizeof(Ulong));

#ifdef DEBUG_CRC
    print("CRC of %d bytes is %d\n",
	  pos[i]+sizeof(Ulong),
	  crc &  255);
#endif    
    walkIndirections(hashBuffer, crc32Buffer,
		     pos, crc, &hc, i+1, -1);
#ifdef DEBUG
    printTable(&hashBuffer, pos, treeDepth);
#endif
  } 
   /* write root-node (write "hc") */
#ifdef DEBUG_CRC
  print("CRC for root-node is %d\n",crc&255);
#endif
  writeRootNode((HashCode160*)&hashBuffer[treeDepth], 
		keywords, num_keys, 
		description, flen, crc);
  /* free rest */
  /* deallocate memory */
  for (i=0;i<treeDepth;i++) 
    free(crc32Buffer[i]); 
  free(crc32Buffer);
  free(hashBuffer);
  free(pos);
  return 0;
}

/* end of insertfile.c */
