/**
 * Values that define the behaviour of GNet that 
 * could be changed without doing much harm. 
 * Users can change these values to fine-tune the
 * behavior of their node. Yet, some of these
 * values can currently not be changed without
 * making the node incompatible with other nodes.
 * Furthermore, most people don't need to worry,
 * and it would be overkill to put them into
 * gnet.conf. 
 * @file gnet/include/gnettypes.h
 * @author Christian Grothoff
 * @author Tzvetan Horozov
 **/

#ifndef CONFIG_H
#define CONFIG_H

/* indirection table size, defaults to 512 entries,
   reduce if you have very little memory, enlarge if you
   start to overflow often. MUST BE A POWER OF 2!!!*/
#define INDIRECTION_TABLE_SIZE 512

/**
 * Number of hosts that can wait for the same query at
 * any given point. Memory cost is
 * INDIRECTION_TABLE_SIZE * 20 * THIS_NUMBER, so it
 * is usually not really costly. Default: 16.
 **/
#define MAX_HOSTS_PER_QUERY 16


/* by which amount do we decrement the TTL for
   simple forwarding / indirection of the query; in seconds.
   Set somewhat in accordance to your network latency (above
   the time it'll take you to send a packet and get a reply). */
#define TTL_DECREMENT 5

/* How many GProxy connections can a single node stand?
   If you expect to have more than one TCP-GProxy client
   at a time, a higher value may be indicated. For
   one-user at a time-settings, 2 should be good enough.
   Don't set it to 1, though. */
#define MAX_GPROXY_CONNECTIONS 2

/* Number of Entries in the on-disk database. Should be larger than the
   number of 1k blocks shared and must be a power of 2. The database-file
   will be 32bytes * CACHE_HASH_TABLE_SIZE. If the table is sparse, the
   database-file may occupy less actual space on the drive. 
   Default is 8388608, which corresponds to a 256 MB database file.
   This is sufficient to share about 8 GB of data (yet, if you get
   closer to this number, efficiency degrades and you may want to
   double the size of this table). You can increase this number up to
   2^31 (= 2 GB database, or 64 GB shared files). If you dedicate more
   space to GNet, we suggest running multiple nodes... */
#define CACHE_HASH_TABLE_SIZE 8388608

/* THE OPTIONS BELOW SHOULD BE TREATED WITH EXTREME CAUTION. 
   THEY *WILL* BREAK COMPATIBILITY WITH ALL OTHER NODES IN THE
   NETWORK. THEY ARE HERE SUCH THAT DEVELOPERS CAN CHANGE
   CONTROVERSIAL SETTINGS. DON'T TOUCH IF YOU DON'T KNOW WHAT
   YOU'RE DOING. PLEASE DO NOT CONNECT TO THE MAIN GNET-NETWORK
   WITH THESE SETTINGS - OTHER NODES WILL IGNORE YOU ANYWAY... */

/* size of the blocks we slice file data into (and indirection node size) */
#define CONTENT_SIZE 1024

/* maximum number of matching files per keyword, default: 1024 */
#define MAX_FSIZE 64

/**
 * The (maximum) size of each message. More is not allowed!
 **/
#define MESSAGE_SIZE 1472

/* max size of a packet - tuned for UDP/Ethernet Framesize (10 MBit) */
#define BUFSIZE (MESSAGE_SIZE-sizeof(MessagePack))

/* minimum size for an immediate flushing of the buffer; tuned for
   a nice milage between response times and traffic analysis (uniform
   packet sizes...) */
#define THRESHOLD 1023

/**
 * Maximum number of replies for the same request/keyword.
 **/
#define MAX_REPLIES 16

/**
 * Granularity of the cron jobs (how precise is the timer,
 * in seconds). 1 is highest, do not set above 120.
 **/
#define CRON_GRANULARITY 1

#endif
/* end of config.h */
