/**
 * The most important types in GNET
 * @file gnet/include/gnettypes.h
 * @author Christian Grothoff
 * @author Tzvetan Horozov
 * @author Ioana Patrascu
 **/

#ifndef GNETTYPES_H
#define GNETTYPES_H

#include <openssl/conf.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/bn.h>

#include <stdarg.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <resolv.h>
#include <zlib.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "semaphore.h"
#include "config.h"

/* *********** CONSTANTS AND TYPES FOR THE ENCRYPTION BUSINESS *********** */

/** type for crc32 results */
#define Ulong uLong

/**
 * Length of RSA encrypted data (2048 bit)
 **/
#define RSA_ENC_LEN 256

/**
 * Length of an RSA KEY (d,e,len), 2048bit key d, 2 byte e, 2 byte len
 **/
#define RSA_KEY_LEN 258

/** length of the sessionkey in bytes
   (128 BIT sessionkey) */
#define SESSIONKEY_LEN (128/8)

/** size of blowfish key in bytes */
#define BF_KEYSIZE 16

/** this BLOCK_LENGTH is 64 bit, blowfish */
#define BLOWFISH_BLOCK_LENGTH 8 

/** the maximum number of pending cron-jobs;
    choose conservatively, doesn't take much
    memory */
#define MAX_CRON_JOBS 8

/**
 * A 160-bit hashcode
 **/
typedef struct {
  int a __attribute__ ((packed));
  int b __attribute__ ((packed));
  int c __attribute__ ((packed));
  int d __attribute__ ((packed));
  int e __attribute__ ((packed));
} HashCode160 __attribute__ ((packed));


/* *********************** Strings - typed ************************** */

typedef unsigned char * FileName;

typedef unsigned char * Interfaces;

typedef unsigned char * IPAddress;

/** type used for length of blocks; except for whole files */
typedef unsigned short BLOCK_LENGTH;

/** size (in bytes) of BLOCK_LENGTH */
#define BLOCK_LENGTH_SIZE sizeof(BLOCK_LENGTH)

/* ******************** Key datastructures ********************** */

typedef struct {
  BLOCK_LENGTH len __attribute__ ((packed));
  BLOCK_LENGTH sizen __attribute__ ((packed));  
  unsigned char key[RSA_KEY_LEN] __attribute__ ((packed));
} PublicKey __attribute__ ((packed));

typedef struct {
  BLOCK_LENGTH len;
  BLOCK_LENGTH sizen;
  BLOCK_LENGTH sizee;
  BLOCK_LENGTH sized;
  BLOCK_LENGTH sizep;
  BLOCK_LENGTH sizeq;
  BLOCK_LENGTH sizedmp1;
  BLOCK_LENGTH sizedmq1;
  unsigned char key[0];
} HostKeyEncoded;

typedef struct {
  unsigned char sig[RSA_ENC_LEN]; 
} Signature;

/* **************** Network protocol business ************************ */

typedef struct {
  BLOCK_LENGTH len __attribute__ ((packed));
  unsigned char encoding[RSA_ENC_LEN] __attribute__ ((packed));
} RSAEncryptedData;

typedef struct {
  unsigned char encoding[0] __attribute__ ((packed));
} BLOWFISHEncryptedData;

/**
 * Constants defining subprotocols (0 <= x <= 65535)
 **/
/** announcement of public key */
#define GNET_PROTO_HELO 0
/** session key exchange, session key is encrypted with hostkey */
#define GNET_PROTO_SKEY 1
/** query for content */
#define GNET_PROTO_QUERY 2
/** receive content */
#define GNET_PROTO_CONTENT 3
/** ack of a session key */
#define GNET_PROTO_SKEY_ACK 4
/** query for content with Triple-hash */
#define GNET_PROTO_3QUERY 5

/** protocol number of UDP. Do NEVER change! */
#define UDP_PROTOCOL_NUMBER 17

/** bit set in the packet size to indicate encryption */
#define ENCRYPTED_FLAG 0x8000


/**
 * Network-identity of a host (IP + Port)
 **/
typedef struct {
  unsigned short senderPort __attribute__ ((packed)); /* claimed port of the sender */
  struct in_addr senderIP __attribute__ ((packed)); /* claimed IP of the sender */  
} HostAddress __attribute__ ((packed));

/**
 * The identity of the host (basically the RIPE160 hashcode of
 * it's public key) - at least at the moment. Treat as opaque!
 **/
typedef struct {
  HashCode160 hashPubKey;
} HostIdentity __attribute__ ((packed));

/**
 * Message-subpart header
 **/
typedef struct {
  /** size of this MESSAGE_PART */
  BLOCK_LENGTH size;
  /** type of the request, GNET_PROTO_XXX */
  unsigned short requestType; 
  /* this is followed by a requestType specific data block,
     consisting of size bytes. */
} MessagePart;

/**
 * Message-Packet header. The packCRC must match the structure
 **/
typedef struct {
  /** size of the message, in bytes, including this header; max 65536-header */
  BLOCK_LENGTH size __attribute__ ((packed));
  /** CRC checksum of the packet */ 
  int checksum __attribute__ ((packed));
  /** Where did the packet come from (IP, port) */  
  HostAddress sender __attribute__ ((packed));
  /** What is the identity of the sender (hash of public key) */
  HostIdentity senderIdentity __attribute__ ((packed));
  
  /* redundant: here is the size of all MESSAGE_PARTs together that follow (type: BLOCK_LENGTH) */
  /* this struct is followed by MESSAGE_PARTs - until size is reached */ 
  /* there is no "end of message" */
  MessagePart parts[0] __attribute__ ((packed));
} MessagePack __attribute__ ((packed));


/** BLOWFISHEncryptedData with enough space for anything we use */
typedef struct {
  unsigned char encoding[BUFSIZE+sizeof(MessagePack)] __attribute__ ((packed));
} BLOWFISHEncryptedDataMax;

/**
 * HELO.
 * It is followed by the actual public key, an inlined
 * array of unsinged chars. The size of that array can
 * be determined using the information in the main header.
 * <br>
 * BEFORE the public key, the SIGNATURE of the host
 * signs the sender-address and a timestamp. A date 
 * should be added to the
 * signed part of the message, indicating how long
 * it is valid (24h, 1 year, etc.). Time should
 * be the date it expires, in UTC.
 **/
typedef struct {
  /** Whose identity follows? No, this is NOT a duplicate
      as a node may send us the identity of ANOTHER node! */
  HostAddress sender __attribute__ ((packed)); 
  HostIdentity senderIdentity __attribute__ ((packed)); 
  /** time this address expires */ 
  time_t expires __attribute__ ((packed));
  /** The signature **/
  Signature signature __attribute__ ((packed));
  /** The public key */
  PublicKey publicKey __attribute__ ((packed)); 
} HELO_Body __attribute__ ((packed));

/**
 * The complete HELO message (still only a part of a packet).
 **/
typedef struct {
  MessagePart header __attribute__ ((packed));
  HELO_Body body __attribute__ ((packed));
} HELO_Message __attribute__ ((packed));

/**
 * The body of a sessionkey-message.
 **/
typedef struct {
  /** the encrypted session key */ 
  RSAEncryptedData key __attribute__ ((packed)); 
  /** time when this key was created */ 
  time_t created __attribute__ ((packed)); 
  /** Signature of the stuff above */
  Signature signature __attribute__ ((packed));
} SKEY_Body __attribute__ ((packed));

/**
 * Session key exchange.
 * The header is followed by an inlined SKS.
 **/
typedef struct {
  MessagePart header __attribute__ ((packed)); 
  SKEY_Body body __attribute__ ((packed));
} SKEY_Message __attribute__ ((packed));

/**
 * Request for content
 * The header is followed by the n hashcodes, each preceeded
 * by its length (as usual)
 **/
typedef struct {
  MessagePart header __attribute__ ((packed)); 
  /** how important is this request */
  int priority __attribute__ ((packed));         
  /** time to live in units of 5s */
  int ttl __attribute__ ((packed));              
  /** to whom to return results? */
  HostIdentity returnTo __attribute__ ((packed));
  /** the file we're looking for */
  HashCode160 query __attribute__ ((packed)); 
} QUERY_Message __attribute__ ((packed));

/**
 * A CONTENT_SIZE block.
 **/
typedef struct {
  unsigned char content[CONTENT_SIZE]; 
} CONTENT_Block __attribute__ ((packed));

/** 
 * type for session keys 
 **/
typedef struct {
  unsigned char key[SESSIONKEY_LEN];
} SESSIONKEY __attribute__ ((packed));

/**
 * Structure for replies to Gproxy.
 **/
typedef struct {
  /** The double-hash */
  HashCode160 hash __attribute__ ((packed));
  /** The content */
  CONTENT_Block content __attribute__ ((packed));
} GPROXY_Reply __attribute__ ((packed));

/**
 * Content (returned).
 **/
typedef struct {
  MessagePart header __attribute__ ((packed));   
  GPROXY_Reply reply __attribute__ ((packed));
} CONTENT_Message __attribute__ ((packed));

/**
 * A structure used to acknowledge the receipt of a SKEY
 **/
typedef struct {
  MessagePart header __attribute__ ((packed));   
  HashCode160 hash __attribute__ ((packed)); // H(E(K))
} ACK_Message __attribute__ ((packed));

/**
 * Structure for the incoming GProxy messages (TCP link!)
 **/
typedef struct {
  /** how important is this request */  
  unsigned int priority __attribute__ ((packed)); 
  /** time to live in seconds */
  unsigned int ttl __attribute__ ((packed));      
  HashCode160 query __attribute__ ((packed));
} GPROXY_Message __attribute__ ((packed));

/**
 * Combined message with HELO and SKEY, with wrapper for
 * sendToHost.
 **/
typedef struct {
  /* length of this struct - 2 */
  HELO_Message helo __attribute__ ((packed));
  SKEY_Message sks __attribute__ ((packed));
} HELOSKEY __attribute__ ((packed));

typedef struct {
  BLOCK_LENGTH len __attribute__ ((packed));
  SESSIONKEY skey __attribute__ ((packed));
} SessionKeyEncoded;

/* ******************** Policy Datastructures *********************** */

/** drop the query if & with this bitmask is 0 */
#define QUERY_DROPMASK 0xF000
/** answer if local files match */
#define QUERY_ANSWER   0x2000
/** forward the query, priority is encoded in QUERY_PRIORITY_BITMASK */
#define QUERY_FORWARD  0x4000
/** forward the query with us as the originating node */
#define QUERY_INDIRECT 0x8000
/** priority to use */
#define QUERY_PRIORITY_BITMASK 0x0FFF

/** type of the results of the polciy module */
typedef unsigned int QUERY_POLICY;

/* *********************** FileFormats ***************************** */

#define MAX_DESC_LEN (CONTENT_SIZE-sizeof(size_t)-sizeof(Ulong)-sizeof(HashCode160)-BLOCK_LENGTH_SIZE)


/**
 * The structure of the root node - contains pertinent information
 * for the file (File length, checksum, hashcode of main indirection
 * node, description length, and description.
 **/
typedef struct {
  size_t file_length __attribute__ ((packed));
  Ulong crc __attribute__ ((packed));
  HashCode160 hashCode __attribute__ ((packed));
  BLOCK_LENGTH descSize __attribute__ ((packed));
  /** followed by description and padding */
  char description[MAX_DESC_LEN] __attribute__ ((packed));
} RootNode __attribute__ ((packed));

/**
 * Type of the content index file entries.
 **/
typedef struct {
  /** the double-hash of this entry */
  HashCode160 doubleHash __attribute__ ((packed));
  /** the current rating of this content */
  unsigned int importance __attribute__ ((packed));
  /** the filename, 0  for entry not used, -1 for data/content shared, > 0 for on-demand encoding */
  unsigned int fileNameIndex __attribute__ ((packed)); 
  /** the offset in the file, offset for on-demand-encoding */
  unsigned int fileOffset __attribute__ ((packed));  
} ContentIndex __attribute__ ((packed));

/**
 * Indirection table entry. Lists what we're looking for,
 * where to forward it, and how long to keep looking for it.
 **/
typedef struct {
  HashCode160 hash; /* what are we looking for? */
  Semaphore * lock; /* lock for this bucket */
  unsigned int ttl; /* when can we forget about this entry? */
  unsigned int priority; /* how much is this query worth? */
  HashCode160 seen[MAX_REPLIES]; /* which replies have we already seen? */
  unsigned int seenIndex; /* at which index should we write the next seen entry? */
  unsigned int hostsWaiting; /* number of hosts waiting for this query */
  HostIdentity destination[MAX_HOSTS_PER_QUERY]; /* where to send it if we get it */
} IndirectionTableEntry;


/* ************************ RUN-TIME Data Structures *********************** */

/**
 * Type of the connection table.
 **/
typedef struct {
  /** semaphore guarding access to this BE */
  Semaphore * sem;
  /** which host are we talking to here? */
  HostIdentity hostId;
  /** what is the credibility of the host? signed
      because that makes many operations that go 
      negative easier. Of course, negative credit makes
      no sense. */
  int credit;
  /** the current session key */
  SESSIONKEY skey;
  /** at which time was the sessionkey created (by whichever party) */
  time_t created;
  /** is this host alive? bitmap marking the times of the last-active intervals */
  unsigned int isAlive; /* 1 true, 0 false, rotated!*/
  /** priority for flushing/sending this buffer, max(priorities of packets in buffer) */
  unsigned int prio; /* what is the max. priority in this buffer */
  /** the buffer containing the data to be send to this host */
  BLOCK_LENGTH len;
  /** contents of the buffer */
  unsigned char * buffer; /* the buffered content, first BLOCK_LENGTH_SIZE bytes give how full the buffer is */
} BufferEntry;

/**
 * Hash2Hex: filename, consisting of 32 bytes [0-9A-Z] plus
 * null-termination.
 **/
typedef struct {
  unsigned char data[sizeof(HashCode160)*2+1];
} HexName;

/**
 * The name of a directory as obtained from the
 * configuration module. Should NEVER be freed!
 **/
typedef struct {
  char dirName[0];
} DirName;

/**
 * Structure to represent n HexNames obtained from
 * scanning a directory.
 **/
typedef struct {
  HexName files[0];
} DirContents;

/**
 * The Delta-list for the cron jobs.
 **/
typedef struct {
  /** The time to wait for this event (in seconds). */
  int delta;
  /** The method to call at that point. */
  void * method;
  /** for cron-jobs: when this should be repeated
      automatically, 0 if this was a once-only job */
  int deltaRepeat;
  /** The index of the next entry in the delta list
      after this one */
  int next;
} DeltaEntry;

/* finally, make sure everybody has the debugging code */
#include "debugging.h"

#endif

/* end of gnettypes.h */
