/**
 * Main handler for incoming packets.
 * @author Christian Grothoff
 * @file gnet/include/handler.h
 **/

#ifndef HANDLER_H
#define HANDLER_H

#include "gnettypes.h"
#include "identity.h"
#include "checksum.h"
#include "keyservice.h"
#include "policy.h"
#include "routing.h"
#include "tcpListener.h"

/**
 * The actual main method of gnet: message dispatch/handling.
 * @param msg the message that was received. caller frees it on return
 * @param id number of the current thread
 **/
void handle(int id, MessagePack * msg);

/**
 * Verify that this message is correct (CRC, sym. encoding).
 * How it works:
 * 0) check policy: do we currently accept anything from that source?
 * 1) IF highest bit in size is set, decrypt:
 *    decrypt packet with key from local database
 *    (senderPort + senderIP => symmetric key)
 *    --- if not encrypted: return 2 at the end!
 *    Important: also the checksum is encoded!
 * 2) check CRC checksum for decrypted packet (rest!)
 *    If wrong, return 0.
 * @returns 0 on error, 1 if ok (and the message was encrypted), 2 if the message is ok but not encrypted
 **/
int verify(MessagePack * msg);

/**
 * Take the HELO message and add the public key to the
 * list of known hosts. Also bind the identity to the
 * connection information provided.
 * @param msg the HELO message
 **/
void handleHELO(HELO_Message * msg);

/**
 * Signed session key exchange.
 * @param pack the header of the message (packet)
 * @param msg the part that contains the new Sessionkey
 **/
void handleSKEY(MessagePack * pack, SKEY_Message * msg);

/**
 * Handle the ack
 */
void handleSKEYACK(MessagePack * pack, ACK_Message * msg);

/**
 * Handle query for content. Depending on how we like
 * the sender, lookup or even forward.
 * @param pack the header of the message (packet)
 * @param msg the part that contains the query
 **/
void handleQUERY(MessagePack * pack,QUERY_Message * msg);

/**
 * Handle triple-hash query for content. Depending on how we like
 * the sender, lookup or even forward.
 * @param pack the header of the message (packet)
 * @param msg the part that contains the query
 **/
void handle3QUERY(MessagePack * pack,QUERY_Message * msg);

/**
 * Receive content, do something with it!
 * There are 3 basic possiblilities. Either our
 * node did the request and we should send the result
 * to gproxy via TCP, or the content was requested by
 * another node and we forwarded the request (and thus
 * we now have to fwd the reply) or 3rd somebody just
 * send us some content we did NOT ask for - and we
 * can choose to store it or just discard it.
 * @param pack the header of the message (packet)
 * @param msg the part that contains the content
 **/
void handleCONTENT(MessagePack * pack, CONTENT_Message * msg);

#endif
/* end of handler.h */
