/**
 * Hostkey - management of the node's main identity.
 * @author Christian Grothoff
 * @file gnet/include/hostkey.h
 **/

/**
 * We currently do not handle encryption of data
 * that can not be done in a single call to the
 * RSA methods (read: large chunks of data).
 * We should never need that, as we can use
 * the hash for larger pieces of data for signing,
 * and for encryption, we only need to encode sessionkeys!
 **/

#ifndef HOSTKEY_H
#define HOSTKEY_H


#include "gnettypes.h"
#include "configuration.h"

#define HOSTKEY RSA *

/**
 * Initialize Random number generator.
 **/
void initRAND(); 

/**
 * create a new hostkey. Callee must free return value.
 **/
HOSTKEY makeHostkey(); 

/**
 * Free memory occupied by hostkey
 * @param hostkey pointer to the memory to free
 **/
void freeHostkey(HOSTKEY hostkey); 

/**
 * Extract the public key of the host.
 * @param result where to write the result.
 **/
void getPublicKey(HOSTKEY hostkey,
		  PublicKey * result);

/**
 * Encode the private key in a format suitable for
 * storing it into a file.
 * @param hostkey the hostkey to use
 * @returns encoding of the private key.
 **/
HostKeyEncoded * encodeHostkey(HOSTKEY hostkey);

/**
 * Decode the private key from the file-format back
 * to the "normal", internal, RSA format.
 * @param encoded the encoded hostkey
 * @returns the decoded hostkey
 **/
HOSTKEY decodeHostkey(HostKeyEncoded * encoding);

/**
 * Encrypt a block with the public key of another
 * host that uses the same cyper.
 * @param block the block to encrypt
 * @param size the size of block
 * @param publicKey the encoded public key used to encrypt
 * @param target where to store the encrypted block
 * @returns 0 on error, 1 if ok
 **/
int encrypt(void * block, 
	    BLOCK_LENGTH size,
	    PublicKey * publicKey,
	    RSAEncryptedData * target);

/**
 * Decrypt a given block with the hostkey. 
 * @param hostkey the hostkey to use
 * @param block the data to decrypt, encoded as returned by encrypt, not consumed
 * @param result pointer to a location where the result can be stored
 * @param max the maximum number of bits to store for the result, if
 *        the decrypted block is bigger, an error is returned
 * @returns the size of the decrypted block, -1 on error
 **/
int decrypt(HOSTKEY hostkey, 
	    RSAEncryptedData * block,
	    void * result,
	    unsigned int max);

/**
 * Sign a given block.
 * @param block the data to sign, first BLOCK_LENGTH_SIZE bytes give length
 * @param size how many bytes to sign
 * @param result where to write the signature
 * @return 0 on error, 1 on success
 **/
int sign(HOSTKEY hostkey, 
	 BLOCK_LENGTH size,
	 void * block,
	 Signature * result);

/**
 * Verify signature.
 * @param block the signed data
 * @param len the length of the block 
 * @param sig signature
 * @param publicKey public key of the signer
 * @returns 1 if ok, 0 if invalid
 **/
int verifySig(void * block,
	      BLOCK_LENGTH len,
	      Signature * sig,	      
	      PublicKey * publicKey);



/* end of hostkey.h */
#endif
