/**
 * This module keeps track of host-identity - address (IP) translation.
 * Basically a mapping HostAddress <==> HostIdentity <==> Public Key 
 * with signature and expiration date for the HostAddress.
 * <p>
 * This file basically is all about the HELO_Message.
 * @file gnet/include/identity.h
 * @author Christian Grothoff
 * @author Tzvetan Horozov
 **/

#ifndef IDENTITY_H
#define IDENTITY_H

#include "gnettypes.h"
#include "checksum.h"
#include "storage.h"
#include "keyservice.h"
#include "sessionkey.h"
#include "semaphore.h"

/** 
 * Initialize identity module. Requries configuration.
 **/
void initAddress();

/**
 * Bind a host identity (IP+port) to a hostId.
 * @param hostId the hostId (hash of publickey)
 * @param conInfo the connection information (IP+port) for the host
 **/
void bindAddress(HELO_Message * msg);

/**
 * Resolve identity (find matching connection)
 * @param hostId the host id
 * @param result the connection information that was bound to that identity or NULL if no information is available
 * @return 0 on failure, 1 on success
 **/
int getConnectionInformation(HostIdentity *  hostId,
			     HostAddress * result);

/**
 * Obtain the identity information for the current node. DO NOT FREE!
 **/
HostAddress * getAddress();

/**
 * Send a packet to a given host.
 * @param hostId the id of the target host
 * @param size the size of the content
 * @param content the actual data (after the header)
 * @param encryptedFlag is the data encrypted (subprotocol)
 * @param crc is the checksum of the unencrypted data (without header)
 **/
void sendToHost(HostIdentity * hostId,
		BLOCK_LENGTH size,
		void * content,
		int encryptedFlag,
                int crc);

/**
 * Obtain the public key of a known host.
 * @param hostId the host id
 * @param result where to store the result
 * @returns 0 on failure, 1 on success
 **/
int identity2Key(HostIdentity *  hostId,
		 PublicKey * result);
 
/**
 * Create the signed HELO message for this host.
 * @param msg the address where to write the HELO message to
 **/
void createHELO(HELO_Message * msg);

/**
 * Verify that a HELO message is still valid.
 * @returns 1 if the message is invalid, 0 if it is ok.
 **/
int verifyHELO(HELO_Message * msg);


#endif
/* end of identity.h */
