/**
 * This module encapsulates the (private) hostkey.
 * @author Christian Grothoff
 * @file gnet/include/keyservice.h
 **/

#ifndef KEYSERVICE_H
#define KEYSERVICE_H

#include "gnettypes.h"
#include "identity.h"
#include "configuration.h"
#include "hostkey.h"
#include "sessionkey.h"
#include "storage.h"
#include "connection.h"

/**
 * Initialize KeyService. Also initializes hostkey's randInit.
 **/
void initKeyService(); 

/**
 * Get the public key of the host
 * @return reference to the public key. Do not free it!
 **/
PublicKey * getPublicHostkey();

/**
 * Obtain identity from publicHostkey.
 * @param pubKey the public key of the host
 * @param result address where to write the identity of the node
 **/
void getHostIdentity(PublicKey * pubKey,
		     HostIdentity * result);

/**
 * Check if 2 hosts are the same (returns 1 if yes)
 * @param first the first host
 * @param second the second host
 * @returns 1 if the hosts are the same, 0 otherwise
 **/
int hostIdentityEquals(HostIdentity * first, 
		       HostIdentity * second);
 
/** 
 * Sign arbitrary data. ALWAYS use only on data we entirely generated.
 * @param data what to sign
 * @param size how big is the data
 * @param result where to store the result
 * @returns 0 on failure, 1 on success
 **/
int signData(void * data,
	     BLOCK_LENGTH size,
	     Signature * result);

/**
 * Decrypt a given block with the hostkey. 
 * @param block the data to decrypt, encoded as returned by encrypt, not consumed
 * @param result pointer to a location where the result can be stored
 * @param max the maximum number of bits to store for the result, if
 *        the decrypted block is bigger, an error is returned
 * @returns the size of the decrypted block, -1 on error
 **/
int decryptData(RSAEncryptedData * block,
		void * result,
		unsigned int max);




/* end of keyservice.h */

#endif
