/**
 * GNET MAIN. This is the server-process that must run on each
 * and every GNET node. The code in here initializes the subsystems
 * and then dispatches incomming UDP packets using multiple threads
 * to the handler(s).
 *
 * @file gnet/node/gnet.c
 * @author Christian Grothoff
 * @author Larry Waldo
 * @author Tzvetan Horozov
 **/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <pthread.h>
 
#include "threads.h"
#include "semaphore.h"
#include "handler.h"
#include "configuration.h"
#include "keyservice.h" 
#include "lookup.h"
#include "statusCalls.h"
#include "startup.h"


/* threads and the shared buffers */
pthread_t GNET_threads_[THREAD_COUNT];

char GNET_buffers[QUEUE_LENGTH][MESSAGE_SIZE];
int GNET_bufferQueue_[QUEUE_LENGTH];
int GNET_bq_firstFree_;
int GNET_bq_lastFree_;
int GNET_bq_firstFull_;

Semaphore * GNET_bufferQueueRead_;
Semaphore * GNET_bufferQueueWrite_;   
Semaphore * GNET_globalLock_;


/**
 * This is the main loop of each thread.
 * It loops *forever* waiting for incomming packets
 * in the packet queue. Then it calls 
 * "handle" (defined in handler.c) on the packet.
 **/
void threadMain(int id) {
  int current;
  int needed = 1;
  
  while (needed) {
    semaphore_down(GNET_bufferQueueRead_);
    /* handle buffer entry */
#ifdef PRINT_THREADS
    print("THREADS: Thread %d handles packet.\n",id);
#endif
    /* sync with other handlers to get buffer */
    semaphore_down(GNET_globalLock_);
    current = GNET_bufferQueue_[GNET_bq_firstFull_++];
    if (GNET_bq_firstFull_ == QUEUE_LENGTH)
      GNET_bq_firstFull_ = 0;
    semaphore_up(GNET_globalLock_);
    /* end of sync */
#ifdef PRINT_THREADS
    print("THREADS: Thread %d handles packet from buffer %d.\n",id,current);
#endif

    /* handle buffer - now out of sync */    
    handle(id,(MessagePack*)GNET_buffers[current]);

    /* free buffer */
    semaphore_down(GNET_globalLock_);
    GNET_bufferQueue_[GNET_bq_lastFree_++] = current;
    if (GNET_bq_lastFree_ == QUEUE_LENGTH)
      GNET_bq_lastFree_ = 0;
    semaphore_up(GNET_globalLock_);
    /* buffer is back in pool. */
#ifdef PRINT_THREADS
    print("THREADS: Thread %d done handling packet.\n",id);
#endif
    /* handling done - open buffer for write */    
    semaphore_up(GNET_bufferQueueWrite_);    
  }
} /* end of threadMain */


/**
 * The main method of GNET. And here is how it works:
 * 1) initialize subsystems 
 *    (configuration, semaphores, key library, output queue)
 * 2) initialize synchronization mechanism for the threads
 * 3) initialize threads, they execute "threadMain"
 *    and query the input queue with incomming packets
 * 4) initialize network (UDP listen)
 * 5) forever: get UDP packet, put it in the queue for the threads
 * 6) [not yet reachable]: close socket on exit
 **/
int main(int argc, char * argv[]) {
  struct sockaddr_in incomming;
  int socket;
  socklen_t addrlen = sizeof(incomming);  
  int size;
  int i;
  int current;
  char * configFile = "~/.gnet\0"; /* default config file */
 
  /* initialize subsystems */
  if (parseCommandLine(argc, argv, &configFile))
    return 0; /* parse error */
  readConfig(configFile);

  checkCompiler();
  initCron();
  initStatusCalls();
  initAddress();
  initRouting();
  initKeyService();
  initConnection(); 
  initLookup();
  initQuery();
  startTCPListener();
  
  /* initialize sync mechanisms for UDP threads */
  GNET_bufferQueueRead_ = new_semaphore(0);
  GNET_bufferQueueWrite_ = new_semaphore(QUEUE_LENGTH);
  GNET_globalLock_ = new_semaphore(1);
  for (i=0;i<QUEUE_LENGTH;i++)
    GNET_bufferQueue_[i]=i;
  GNET_bq_firstFree_ = 0;
  GNET_bq_lastFree_ = 0;
  GNET_bq_firstFull_ = 0;
  
  /* create UDP threads */
  for (i=0;i<THREAD_MIN;i++) 
    pthread_create(&GNET_threads_[i], NULL,
		   (void * (*)(void *)) threadMain,
		   (void *) &i);  

  /* initialize network */
  bzero(&incomming,sizeof(struct sockaddr_in));
  socket = passivesock();

  /* start control thread */
  startCron();

  /* listen & distribute UDP packets to threads */
  while (1) {
    semaphore_down(GNET_bufferQueueWrite_);
    /* aquire buffer */
    semaphore_down(GNET_globalLock_);
    if (GNET_bq_firstFree_ == QUEUE_LENGTH)
      GNET_bq_firstFree_ = 0;
    current = GNET_bufferQueue_[GNET_bq_firstFree_++];    
    semaphore_up(GNET_globalLock_);
#ifdef PRINT_UDP
    print("UDP: Receiving from socket %d, buffer %d.\n",
	  socket, current);
#endif    
    size = recvfrom(socket, GNET_buffers[current], MESSAGE_SIZE, 0,
		    (struct sockaddr * )&incomming, &addrlen);
    /* quick test of the packet, if failed, repeat! */
    while ((size != ((((MessagePack*)GNET_buffers[current])->size+BLOCK_LENGTH_SIZE) & 
		     (ENCRYPTED_FLAG-1))) ||
	   (size < sizeof(MessagePack))) {
#ifdef PRINT_WARNINGS
      print("Received invalid packet (%d != %d). Discarded.\n",
	    size,
	    (((MessagePack*)GNET_buffers[current])->size+BLOCK_LENGTH_SIZE)
	    & (ENCRYPTED_FLAG-1));
#endif     
      size = recvfrom(socket, GNET_buffers[current], MESSAGE_SIZE, 0,
		      (struct sockaddr * )&incomming, &addrlen);
    }
#ifdef PRINT_UDP
    print("UDP: Received %d bytes from socket %d, buffer %d.\n",
	  size,socket, current);
#endif    
   /* finished with list */
    semaphore_up(GNET_bufferQueueRead_);
    bzero(&incomming,sizeof(struct sockaddr_in));
  }
  /* shutdown - this can never happen ... */
  close(socket);
  semaphore_destroy(GNET_globalLock_);
  semaphore_destroy(GNET_bufferQueueRead_);
  semaphore_destroy(GNET_bufferQueueWrite_);
  return 0;
} 

/* You have reached the end of GNet. You can shutdown your
   computer now and get a live. */
