/**
 * Hostkey - management of the node's main identity.
 * @author Christian Grothoff
 * @file primitives/hostkey.c
 **/

/**
 * FIXME:
 * + the "random" seed is not set, or is it?
 *   (check with openSSL docs)
 **/

#include "hostkey.h"

#define HOSTKEY_LEN 2048

/**
 * Initialize Random number generator.
 **/
void initRAND() {
  srand((unsigned int)time(NULL));
  /* this is probably not the right random generator for openssl ... */
}

/**
 * This HostKey implementation uses RSA.
 **/
HOSTKEY makeHostkey() {
  return RSA_generate_key(HOSTKEY_LEN,65535,NULL,0);
}

/**
 * Free memory occupied by hostkey
 **/
void freeHostkey(HOSTKEY hostkey) {
  RSA_free(hostkey);
}

/**
 * Extract the public key of the host.
 * @param result where to write the result.
 **/
void getPublicKey(HOSTKEY hostkey,
		  PublicKey * result) {
  BLOCK_LENGTH sizen;
  BLOCK_LENGTH sizee;
  BLOCK_LENGTH size;
  sizen = BN_num_bytes(hostkey->n);
  sizee = BN_num_bytes(hostkey->e);
  size = sizen + sizee+2*BLOCK_LENGTH_SIZE;
  if (size != sizeof(PublicKey))
    errexit("oops: sizeof public key does not match size (%d!=%d)",
	    size,sizeof(PublicKey));
  if (RSA_KEY_LEN != sizen+sizee) 
    errexit("PublicKey datastructure wrong (%d+%d!=%d!",
	    sizen,sizee,RSA_KEY_LEN);
  result->len = size-BLOCK_LENGTH_SIZE;
  result->sizen = sizen;
  BN_bn2bin(hostkey->n,
	    &result->key[0]);
  BN_bn2bin(hostkey->e,
	    &result->key[sizen]);
}

/**
 * Internal: publicKey => RSA-Key
 **/
HOSTKEY public2Hostkey(PublicKey * publicKey) {
  HOSTKEY result = RSA_new();
  int sizen;
  int sizee;
  if (publicKey->len < 200) { /* the real key must be bigger... */
#ifdef PRINT_WARNINGS
    print("public2Hostkey: received invalid publicKey (size=%d)\n",
	  publicKey->len);
#endif
    return NULL;
  }
  sizen = publicKey->sizen;
  sizee = publicKey->len - sizen - BLOCK_LENGTH_SIZE;
  if ((sizee <= 0) || (sizen <= 0) || (sizee + sizen != RSA_KEY_LEN)) {
#ifdef PRINT_WARNINGS
    print("public2Hostkey: received invalid publicKey (sizee=%d, sizen=%d)\n",
	  sizee,sizen);
#endif
    return NULL;
  }
  result->n = BN_bin2bn(&publicKey->key[0], sizen, NULL);
  result->e = BN_bin2bn(&publicKey->key[sizen],sizee, NULL);
  return result;
}

/**
 * Encode the private key in a format suitable for
 * storing it into a file.
 * @returns encoding of the private key.
 *    The first 4 bytes give the size of the array, as usual.
 **/
HostKeyEncoded * encodeHostkey(HOSTKEY hostkey) {
  /*
               BIGNUM *n;               public modulus
               BIGNUM *e;               public exponent
               BIGNUM *d;               private exponent
               BIGNUM *p;               secret prime factor
               BIGNUM *q;               secret prime factor
               BIGNUM *dmp1;            d mod (p-1)
               BIGNUM *dmq1;            d mod (q-1)
               BIGNUM *iqmp;            q^-1 mod p
  */
  BLOCK_LENGTH sizen;
  BLOCK_LENGTH sizee;
  BLOCK_LENGTH sized;
  BLOCK_LENGTH sizep;
  BLOCK_LENGTH sizeq;
  BLOCK_LENGTH sizedmp1;
  BLOCK_LENGTH sizedmq1;
  BLOCK_LENGTH sizeiqmp;
  BLOCK_LENGTH size;
  HostKeyEncoded * retval;
  sizen = BN_num_bytes(hostkey->n);
  sizee = BN_num_bytes(hostkey->e);
  sized = BN_num_bytes(hostkey->d);
  sizep = BN_num_bytes(hostkey->p);
  sizeq = BN_num_bytes(hostkey->q);
  sizedmp1 = BN_num_bytes(hostkey->dmp1);
  sizedmq1 = BN_num_bytes(hostkey->dmq1);
  sizeiqmp = BN_num_bytes(hostkey->iqmp);
  size = sizen+sizee+sized+sizep+sizeq+sizedmp1+sizedmq1+sizeiqmp+sizeof(HostKeyEncoded);
  retval = (HostKeyEncoded *) malloc(size);
  retval->len = size-BLOCK_LENGTH_SIZE;
  retval->sizen = sizen;
  retval->sizee = sizee;
  retval->sized = sized;
  retval->sizep = sizep;
  retval->sizeq = sizeq;
  retval->sizedmp1 = sizedmp1;
  retval->sizedmq1 = sizedmq1;
  BN_bn2bin(hostkey->n,&retval->key[0]);
  BN_bn2bin(hostkey->e,&retval->key[0+sizen]);
  BN_bn2bin(hostkey->d,&retval->key[0+sizen+sizee]);
  BN_bn2bin(hostkey->p,&retval->key[0+sizen+sizee+sized]);
  BN_bn2bin(hostkey->q,&retval->key[0+sizen+sizee+sized+sizep]);
  BN_bn2bin(hostkey->dmp1,&retval->key[0+sizen+sizee+sized+sizep+sizeq]);
  BN_bn2bin(hostkey->dmq1,&retval->key[0+sizen+sizee+sized+sizep+sizeq+sizedmp1]);
  BN_bn2bin(hostkey->iqmp,&retval->key[0+sizen+sizee+sized+sizep+sizeq+sizedmp1+sizedmq1]);
  return retval;
}

/**
 * Decode the private key from the file-format back
 * to the "normal", internal format.
 **/
HOSTKEY decodeHostkey(HostKeyEncoded * encoding) {
  BLOCK_LENGTH sizen;
  BLOCK_LENGTH sizee;
  BLOCK_LENGTH sized;
  BLOCK_LENGTH sizep;
  BLOCK_LENGTH sizeq;
  BLOCK_LENGTH sizedmp1;
  BLOCK_LENGTH sizedmq1;
  BLOCK_LENGTH size;
  BLOCK_LENGTH sum;
  HOSTKEY result = (HOSTKEY) RSA_new();
  size    = encoding->len + BLOCK_LENGTH_SIZE - sizeof(HostKeyEncoded);
  sizen   = encoding->sizen;
  sizee   = encoding->sizee;
  sized   = encoding->sized;
  sizep   = encoding->sizep;
  sizeq   = encoding->sizeq;
  sizedmp1= encoding->sizedmp1;
  sizedmq1= encoding->sizedmq1;
  sum = 0;
  result->n= BN_bin2bn(&encoding->key[sum], sizen, NULL); sum += sizen;
  result->e= BN_bin2bn(&encoding->key[sum], sizee, NULL); sum += sizee;
  result->d= BN_bin2bn(&encoding->key[sum], sized, NULL); sum += sized;
  result->p= BN_bin2bn(&encoding->key[sum], sizep, NULL); sum += sizep;
  result->q= BN_bin2bn(&encoding->key[sum], sizeq, NULL); sum += sizeq;
  result->dmp1= BN_bin2bn(&encoding->key[sum], sizedmp1, NULL); sum += sizedmp1;
  result->dmq1= BN_bin2bn(&encoding->key[sum], sizedmq1, NULL); sum += sizedmq1;
  result->iqmp= BN_bin2bn(&encoding->key[sum], size-sum, NULL);
  return result;
}

/**
 * Encrypt a block with the public key of another
 * host that uses the same cyper.
 * @param block the block to encrypt
 * @param size the size of block
 * @param publicKey the encoded public key used to encrypt
 * @param target where to store the encrypted block
 * @returns 0 on error, 1 if ok
 **/
int encrypt(void * block, 
	    BLOCK_LENGTH size,
	    PublicKey * publicKey,
	    RSAEncryptedData * target) {
  HOSTKEY foreignkey = public2Hostkey(publicKey);
  int rs = RSA_size(foreignkey);
  /* now encrypt. First get size of the block */
  if (size > (rs - 41)) {
#ifdef PRINT_WARNINGS
    print("HostKey::encrypt() called with %d bytes where foreignkey allows only %d\n",
	  size, 
	  rs-41);
#endif
    return 0;
  }
  target->len = RSA_public_encrypt(size, 
				   block, 
				   &target->encoding[0], 
				   foreignkey,
				   RSA_PKCS1_OAEP_PADDING);
  assert(target->len == RSA_ENC_LEN,
	 "RSA-Encoding has wrong length!");
  freeHostkey(foreignkey);
  return 1;
}

/**
 * Decrypt a given block with the hostkey. 
 * @param block the data to decrypt, encoded as returned by encrypt
 * @param result pointer to a location where the result can be stored
 * @param max the maximum number of bits to store for the result, if
 *        the decrypted block is bigger, an error is returned
 * @returns the size of the decrypted block, -1 on error
 **/
int decrypt(HOSTKEY hostkey, 
	    RSAEncryptedData * block,
	    void * result,
	    unsigned int max) {
  RSAEncryptedData tmp; /* this is as big as the result can possibly get */
  int size;

  if (block == NULL)
    return -1;
  size = RSA_private_decrypt(block->len, 
			     &block->encoding[0],
			     &tmp.encoding[0], 
			     hostkey,
			     RSA_PKCS1_OAEP_PADDING);
  if ((size == -1) || (size > max))
    return -1;
  memcpy(result,
	 &tmp.encoding[0],
	 size);
  return size;
}

/**
 * Sign a given block.
 * @param block the data to sign
 * @param size how many bytes to sign
 * @param result where to write the signature
 * @return 0 on error, 1 on success
 **/
int sign(HOSTKEY hostkey, 
	 BLOCK_LENGTH size,
	 void * block,
	 Signature * result) {
#ifdef EXTRA_CHECKS
  PublicKey pkey;
#endif
  int rs = RSA_size(hostkey);
  int sigSize;

  if (block == NULL)
    return 0;
  assert(rs == sizeof(Signature),
	 "Signature length (RSA_size) has unexpected value!");
  sigSize = RSA_private_encrypt(size, /* size */
				(unsigned char *)block, /* data to sign */
				&result->sig[0],
				hostkey,
				RSA_PKCS1_PADDING); /* or: RSA_NO_PADDING? */
  assert(sigSize == sizeof(Signature),
	 "sign: sigSize wrong!?");
#ifdef DEBUG
  printf("Signature size= %d\n",sigSize);
#endif
#ifdef EXTRA_CHECKS
  getPublicKey(hostkey, &pkey);
  assert(1==verifySig(block, size, result, &pkey),
	 "sign: generated signature does not pass verification!");
#endif
  return 1;
}

/**
 * Verify signature.
 * @param block the signed data
 * @param len the length of the block 
 * @param sig signature
 * @param publicKey public key of the signer
 * @returns 1 if ok, 0 if invalid
 **/
int verifySig(void * block,
	      BLOCK_LENGTH len,
	      Signature * sig,	      
	      PublicKey * publicKey) {
  HOSTKEY hostkey;
  int rs,size;
  Signature tmp;

  hostkey = public2Hostkey(publicKey);
  if ((hostkey == NULL) || (sig == NULL) || (block == NULL))
    return 0; /* hey, no data !? */
  rs = RSA_size(hostkey);
  assert(rs == RSA_ENC_LEN,
	 "rs != RSA_ENC_LEN!");
  size = RSA_public_decrypt(RSA_ENC_LEN, /* size */
			    &sig->sig[0], /* signature */
			    &tmp.sig[0],
			    hostkey,
			    RSA_PKCS1_PADDING);
  if (rs < size) 
    errexit("rs < size (%d - %d). Fatal error (in OpenSSL?).\n",
	    rs,size);
  freeHostkey(hostkey);

  if (size != len) {
#ifdef PRINT_WARNINGS
    print("VerifySig: signature mismatch: %d - %d\n",rs,len); 
#endif
    return 0;
  }
  if ((size=memcmp((unsigned char*)block,
		   &tmp.sig[0],
		   size)) != 0) {
#ifdef PRINT_WARNINGS
    print("verifySig: signature mismatch: %d!\n",size); 
#endif
    return 0;   
  }       
  return 1;
}



/* end of HostKey */
