/**
 * Policy implementation.
 * @author Christian Grothoff
 * @file node/policy.c
 **/

#include "policy.h"
#include "storage.h"
#include "configuration.h"
#include "statusCalls.h"

/**
 * A new packet is supposed to be send out. Should it be
 * dropped because the load is too high?
 * <p>
 * @param priority the highest priority of contents in the packet
 * @returns 1 if the packet should be handled, 0 if the packet should be dropped.
 **/
int outgoingCheck(int priority) {
  unsigned int maxNetLoad=getMaxNetLoad();
  unsigned int maxLoad=getMaxLoad();
  unsigned int curNetLoad = networkUsage();
  unsigned int curLoad = cpuUsage();

#ifdef LOAD_DEBUG
  print("CurNetLoad=%d , CurCPULoad=%d \n",
	networkUsage(), cpuUsage());
#endif
  priority++;
  if (curNetLoad*2>maxNetLoad*priority) {
#ifdef PRINT_POLICY
    print("POLICY: network load to high, dropping outgoing.\n");
#endif
    return 0;
  }
  if (curLoad*2>maxLoad*priority) {
#ifdef PRINT_POLICY
    print("POLICY: CPU load to high (%d*2>%d*%d), dropping outgoing.\n",
	  curLoad, maxLoad, priority);
#endif
    return 0;
  }
  return 1; /* no policy */
}

/**
 * A new packet is coming in. The question is, if we should
 * have a look at it at all (incl. decryption). The sender
 * is known. 
 * <p>
 * The policy must decide if we should drop the packet. The
 * reason may be that we don't trust the host or that the load
 * is to high (or both or a combination). 
 * <p>
 * @param senderIdentity the identity of the sender
 * @returns 1 if the packet should be handled, 0 if the packet should be dropped.
 **/
int incomingCheck(HostIdentity * senderIdentity) {
  unsigned int maxNetLoad=getMaxNetLoad();
  unsigned int maxLoad=getMaxLoad();
  unsigned int curNetLoad = networkUsage();
  unsigned int curLoad = cpuUsage();

#ifdef LOAD_DEBUG
  printf("CurNetLoad=%d , CurCPULoad=%d \n",networkUsage(), cpuUsage());
#endif
  
  if (curNetLoad>maxNetLoad) {
#ifdef PRINT_POLICY
    print("POLICY: network load to high (%d > %d), dropping incoming packet.\n",
	  curNetLoad, maxNetLoad);
#endif
    return 0;
  }
  if (curLoad>maxLoad) {
#ifdef PRINT_POLICY
    print("POLICY: CPU load to high (%d > %d), dropping incoming packet.\n",
	  curLoad, maxLoad);
#endif
    return 0;
  }
  if ( (curLoad*2>maxLoad) ||
       (curNetLoad*2>maxNetLoad) ) 
    if (changeHostCredit(senderIdentity,-1) != -1)
      return 0; /* drop: busy and no credit! */  
  return 1; /* no policy */
}

/**
 * A query has been received. The question is, if it should be forwarded
 * and if with which priority. Routing decisions(to whom) are to be taken elsewhere.
 * <p>
 * @param senderIdentity the host *forwarding* (or originating, we don't know) us the query
 * @param priority the priority the query had when it came in, may be an arbitrary number if the 
 *        sender is malicious! Cap by trustlevel first!
 * @return binary encoding: QUERY_XXXX constants
 **/
QUERY_POLICY evaluateQuery(HostIdentity * senderIdentity, 
			   unsigned int priority) {
  unsigned int maxNetLoad=getMaxNetLoad();
  unsigned int maxLoad=getMaxLoad();
  unsigned int curNetLoad = networkUsage();
  unsigned int curLoad = cpuUsage();
  if ( (maxNetLoad > curNetLoad*2) &&
       (maxLoad > curLoad*2) )
    return 1 /* minimum priority */ |
      QUERY_ANSWER | QUERY_FORWARD | QUERY_INDIRECT;
  priority = 1-changeHostCredit(senderIdentity, -priority);
  if (maxNetLoad*2*(priority+1)>curNetLoad*3)
    return ((priority) & QUERY_DROPMASK) |
      QUERY_ANSWER | QUERY_FORWARD | QUERY_INDIRECT;
  else if (maxLoad*2*(priority+1)>(curLoad*3))
    return ((priority) & QUERY_DROPMASK) |
      QUERY_ANSWER | QUERY_FORWARD;
  else if (maxNetLoad*(priority+1)>(curLoad))
    return ((priority) & QUERY_DROPMASK) |
      QUERY_ANSWER;
  else
    return 0; /* drop entirely */
}

/**
 * Some content dropped by. We may want to store it locally, or not.
 * The policy can do with it whatever it sees fit.
 *
 * The current primitive implementation stores content that flys
 * by until the drive is full, disregarding priorities of the
 * requests or the importance of the content already stored.
 * Content is never discarded.
 *
 * This is of course a serious tuning problem... 
 *
 * @param senderIdentity the sender of the content
 * @param content the data
 * @param preference how good was the data?
 **/
void evaluateContent(HostIdentity * senderIdentity,
		     CONTENT_Block * content,
		     HashCode160 * hash,
		     int preference) {
  ContentIndex newEntry;
  HexName fil;

  if (!(spaceAvailable(preference)))
    return;
  hash2hex(hash, &fil);
  writeContent(&fil, content);
  memcpy(&newEntry.doubleHash,
	 hash,
	 sizeof(HashCode160));
  newEntry.importance = preference;
  newEntry.fileNameIndex = -1;
  newEntry.fileOffset = -1;
  addEntry(&newEntry);
}


/* end of policy.c */
