/**
 * Routing implementation.
 * @author Christian Grothoff
 * @file node/routing.c
 **/

#include "routing.h"

IndirectionTableEntry * ROUTING_indTable_;


/* ************* internal Methods **************** */


/**
 * Cron job, updating TTLs in the routing table.
 * Called every TTL_DECREMENT seconds.
 **/
void cronRouting();
 

/* ******************** CODE ********************* */

/**
 * initialize routing module (starts TCP server,
 * initializes indirection table)
 **/
void initRouting() {
  unsigned int i;

  ROUTING_indTable_ 
    = malloc(sizeof(IndirectionTableEntry)*
	     INDIRECTION_TABLE_SIZE);
  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    ROUTING_indTable_[i].ttl = 0; /* expired / free */  
    ROUTING_indTable_[i].hostsWaiting = 0; /* expired / free */  
    ROUTING_indTable_[i].lock = new_semaphore(1);    
  }
  addCronJob(&cronRouting, TTL_DECREMENT ,TTL_DECREMENT);
}

void printIndirectionTable(char * sign) {
#ifdef VERY_VERBOSE
  int i;
  IndirectionTableEntry * ite;
  HexName h1;
  HexName h2;

  print("INDIRECTION TABLE at %d ENTRIES %s: \n",
	(int) ROUTING_indTable_, sign);
  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    ite = &ROUTING_indTable[i];  
    semaphore_down(ite->lock); 
    hash2hex(&ite->hash, &h1);
    hash2hex(&ite->destination[0].hashPubKey, &h2);
    print("%d: hash %s, destination[0] %s, ttl %d, hostsWaiting %d prio %d seenIndex: %d\n ", 
	  &h1, &h2,
	  ite->ttl,
	  ite->hostsWaiting,
	  ite->priority,
	  ite->seenIndex);
    semaphore_up(ite->lock);  
  }
#endif
}

/**
 * Cron job, updating TTLs in the routing table.
 * Called every TTL_DECREMENT seconds.
 **/
void cronRouting(){
  unsigned int i;
  IndirectionTableEntry * ite;

  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    ite = &ROUTING_indTable_[i];
    semaphore_down(ite->lock); 
    if (ite->ttl > 0) {
#ifdef PRINT_ROUTING
      print("ROUTING: Decreasing ttl entry %d from %d to %d.\n",
	    i, ite->ttl,
	    ite->ttl-TTL_DECREMENT );
#endif
     if (ite->ttl < TTL_DECREMENT )
       ite->ttl = 0;
     else 
       ite->ttl -= TTL_DECREMENT ;  
    }   
    semaphore_up(ite->lock); 
  }
}

/**
 * Find out, if this query is already pending, if yes,
 * return 1, otherwise return 0.
 * @param msg the Query to look for
 * @returns 1 if the query was already send out, 0 if not
 **/
int isPending(QUERY_Message * msg) {
  HashCode160 hc;
  IndirectionTableEntry * ite;

#ifdef PRINT_QUERY
  HexName hex;

  print("QUERY: Checking if query is pendnig.\n");
#endif
  if (msg->header.requestType == GNET_PROTO_QUERY) {
    /* convert from double to triple-hash */
    hash(&msg->query,
	 sizeof(HashCode160),
	 &hc); /* better not use aliases... */
    memcpy(&msg->query,
	   &hc,
	   sizeof(HashCode160));
    msg->header.requestType = GNET_PROTO_3QUERY;
  }
  ite = &ROUTING_indTable_[computeRoutingIndex(&msg->query)];
  semaphore_down(ite->lock); 
  if (ite->ttl < msg->ttl) {
    semaphore_up(ite->lock); 
    return 0; /* if there, ttl smaller */
  }
  if (equalsHashCode160(&msg->query,
			&ite->hash)) {
#ifdef PRINT_QUERY
    hash2hex(&ite->hash, &hex);
    print("QUERY: Query is pending %s (with ttl old %d and new %d).\n",
	  &hex,
	  ite->ttl, msg->ttl);
#endif
    semaphore_up(ite->lock); 
    return 1; /* already there, ttl larger */    
  }
  semaphore_up(ite->lock); 
#ifdef PRINT_QUERY
  print("QUERY: Query is new.\n");
#endif
  return 0;
}

/**
 * We're going to forward an external request with
 * indirection. Thus this
 * node must keep track of the request for a while to
 * route results. This method should not call 
 * forwardQuery, the caller will.
 * @param msg the query we're going to forward
 **/
void indirectingQuery(QUERY_Message * msg) {
  int destIndex;
  HashCode160 hc;
  IndirectionTableEntry * ite;
#ifdef PRINT_QUERY
  HexName hex;
#endif
 
  if (msg->ttl <= 0)
    return;
#ifdef PRINT_QUERY
  hash2hex(&msg->query,&hex);
  print("QUERY: Indirecting Query %s.\n",
	&hex);
#endif
  if (msg->header.requestType == GNET_PROTO_QUERY) {
    /* convert from double to triple-hash */
    hash(&msg->query,
	 sizeof(HashCode160),
	 &hc); /* better not use aliases... */
    memcpy(&msg->query,
	   &hc,
	   sizeof(HashCode160));
    msg->header.requestType = GNET_PROTO_3QUERY;
  }

  ite = &ROUTING_indTable_[computeRoutingIndex(&msg->query)];
  semaphore_down(ite->lock);

  if (equalsHashCode160(&msg->query,
			&ite->hash) &&
      (ite->ttl > 0)) {
    /* we may have this host already in the waiting list,
       we definitely have the query pending. Decide between
       doing nothing and appending host to queue */      
    for (destIndex=0;destIndex<ite->hostsWaiting;destIndex++) {
      if (equalsHashCode160(&ite->destination[destIndex].hashPubKey,
			    &msg->returnTo.hashPubKey)) {
#ifdef PRINT_QUERY
	print("QUERY: The query we are indirecting is already in the table.\n");
#endif	
	semaphore_up(ite->lock); 
	return; /* already in routing table! */
      }
    }
    /* Query matches but host is not in the list.
       If possible, we add the host to the
       destinations for the result */
    if (ite->hostsWaiting < MAX_HOSTS_PER_QUERY) 
      memcpy(&ite->destination[ite->hostsWaiting++],
	     &msg->returnTo,
	     sizeof(HostIdentity));
#ifdef PRINT_WARNINGS
    else 
      print("You may want to increase MAX_HOSTS_PER_QUERY.\nThe current limit %d was hit.\n",
	    MAX_HOSTS_PER_QUERY);
#endif    
    semaphore_up(ite->lock); 
    return; /* query pre-existed */
  }
  /* we had no or a different query before. If the old
     query has a smaller tll we keep it, otherwise we
     take the new one. */
  if ( (ite->ttl < msg->ttl) &&
       (ite->ttl > 0) ) {
    /* life query with smaller tll has this slot,
       we can't keep track of this indirection. 
       FIXME: Signaling an error such that we drop the packet
       would probably be a good idea. (Do not forward query
       without indirection, that might make the forward/not-
       forward policy manipulatable from outside). */   
    semaphore_up(ite->lock); 
    return;
  }
  /* Ok, this entry is empty or requires us to keep track
     longer; copy the new query over */
  ite->ttl = msg->ttl;
  ite->priority = msg->priority;  
  ite->hostsWaiting = 1; 
  memcpy(&ite->destination[0],
	 &msg->returnTo,
	 sizeof(HostIdentity));
  memcpy(&ite->hash,
	 &msg->query,
	 sizeof(HashCode160));
  ite->seenIndex = 0;
  semaphore_up(ite->lock); 
}

/**
 * Content has arrived. We must decide if we want to
 * a) forward it to our GProxy
 * b) indirect it to other nodes
 * The routing module should know what to do.
 * @returns how good this content was (priority of the request)
 **/
int useContent(HostIdentity * hostId,
		CONTENT_Block * content,
		HashCode160 * hc) {
  unsigned int i;
  HashCode160 tripleHash;
  IndirectionTableEntry * ite;
  int prio = -1;
  
#ifdef PRINT_CONTENT
  HexName hex;
  hash2hex(hc, &hex);
  print("CONTENT: Using Content %s.\n",
	&hex);
#endif
  hash(hc,
       sizeof(HashCode160),
       &tripleHash);

  ite = &ROUTING_indTable_[computeRoutingIndex(&tripleHash)];
  semaphore_down(ite->lock);
 
  if ( (ite->ttl == 0) ||
       (!equalsHashCode160(&ite->hash,
			   &tripleHash)) ) {	
    semaphore_up(ite->lock);
    return 0; /* no indirection pending */
  }
  for (i=0;i<ite->seenIndex;i++) 
    if (equalsHashCode160(hc,
			  &ite->seen[i])){
#ifdef PRINT_ROUTING
      print("ROUTING: reply already seen, not forwarded.\n");
#endif
      semaphore_up(ite->lock); 
      return 0; /* already forwarded! */	  
    }  
#ifdef PRINT_ROUTING
  hash2hex(&ROUTING_indTable_[i].hash,&hex);
  print("ROUTING: Indirecting CONTENT to %s.\n",
	&hex);
#endif	
  /* new reply, adjust credits! */
  changeHostCredit(hostId, ite->priority);
  prio = ite->priority;
  ite->priority = 1; /* reduce priority for further replies, as we didn't get paid for those... */

  for (i=0;i<ite->hostsWaiting;i++)
    sendReply(&ite->destination[i],
	      content,
	      hc);
  if (ite->seenIndex == MAX_REPLIES)
    ite->ttl = 0; 
  else /* mark as already forwarded */
    memcpy(&ite->seen[ite->seenIndex++],
	   hc,
	   sizeof(HashCode160));
  semaphore_up(ite->lock);
#ifdef PRINT_CONTENT_ALL
  print("CONTENT: Done using Content.\n");
#endif
  return prio;
}


/**
 * Compute the hashtable index of a host id.
 **/
unsigned int computeRoutingIndex(HashCode160 * query) {
  unsigned int res = (((unsigned int)query->a) & 
		      ((unsigned int)(INDIRECTION_TABLE_SIZE - 1)));
#ifdef EXTRA_CHECKS
  if (res >= INDIRECTION_TABLE_SIZE)
    errexit("FATAL: INDIRECTION_TABLE_SIZE not power of 2? (%d)\n",
	    INDIRECTION_TABLE_SIZE);
#endif
  return res;
}


/* end of routing.c */
