/** 
 * Status calls implementation.
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 * @file node/statusCalls.c
 **/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <dirent.h>
#include "statusCalls.h"
#include "configuration.h"

char * interfacePtrs[MAX_INTERFACES]; /* Pointer to each interface */
int numInterfaces = 0;


/**
 * The following method is called in order to initialize the status calls
 * routines. After that it is safe to call each of the status calls separately
 * @return 1 on success and -1 on error.
 **/
int initStatusCalls(){
  Interfaces interfaces=getNetInterfaces();
  char *ch=interfaces;
  int start=-1, done=-1;
#ifdef PRINT_UDP
  int i;
#endif

  /* fail gracefully if config-file is incomplete */
  if (interfaces == NULL) 
    errexit("no network interfaces defined!\n");

  /* The string containitn the interfaces is formatted in the following way: 
   * each comma is replaced by '\0' and the pointers to the beginning of every 
   * interface are stored
   */
  while (done<0){
    if (*ch=='\0'){
      done=1;      
      if (start>0){	
	numInterfaces++;
      }	      
    }
    else if ((*ch>='a'&&*ch<='z') || (*ch>='A'&&*ch<='Z') || (*ch>='0'&&*ch<='9')){
      if (start<0){
	start=1;   
	interfacePtrs[numInterfaces]=ch;	
      }
    }
    else {
      if (start>0){
	start=-1;
	*ch='\0';	
	numInterfaces++;
      }    
    }
    ch++;
  }
  
  if (numInterfaces<=0)
    errexit("ERROR: No network interfaces specified in he configuration file\n");
  if (numInterfaces> MAX_INTERFACES)
    errexit("ERROR: Too many network interfaces specified in the configuration file. Must be less than %d\n",MAX_INTERFACES);


#ifdef PRINT_UDP  
  print("Total interfaces=%d\n",numInterfaces);    
  for (i=0;i<numInterfaces;i++){
    print(" -> %s\n",(char*)interfacePtrs[i]);
  }
#endif

  /* Call the three methods once to initialize the sequence */
  networkUsage(); 
  diskUsage();
  cpuUsage();
  return 1;  
}

/**
 * The following routine returns the percentage of
 * available used bandwidth.
 * Example: If 81 is returned this means that 81% of the network 
 * bandwidth of the host is consumed.
 * The method initStatusCalls() should be called before this routine is 
 * invoked
 * If there is an error the method returns -1
 **/
int networkUsage() {
  static long lastnettime=0; /* In seconds */
  static int lastnetresult=-1;
  static int firstnetcall=0;
  static int last_net_results[MAX_INTERFACES][4];

  FILE* proc_net_dev;
  char line[5000];
  char *data;
  int rxnew, txnew, collnew, found=0, rxdiff=0, txdiff=0, i=0;
  int ifnum=0;
  struct timeval t1;
  long now, elapsedtime;
  char curInterface[255];

  /* Get the current time in microseconds */
  gettimeofday(&t1,NULL);  
  now=t1.tv_sec;
  elapsedtime=now-lastnettime;
  if ((elapsedtime < 2) && (lastnetresult != -1))
    return lastnetresult;
  lastnettime=now;

  proc_net_dev = fopen("/proc/net/dev", "r");
  /* Try to open the file*/
  if (!proc_net_dev) {
    print("ERROR: Could not open /proc/net/dev (%d)\n", proc_net_dev);
    return 0;
  }
 
  /* Parse the line with 'eth0' */
  while (!feof(proc_net_dev)&& ifnum<numInterfaces) {    
    fgets(line, 5000, proc_net_dev);
    
    for (i=0;i<numInterfaces;i++){
      strcpy(curInterface,interfacePtrs[i]);
      strcat(curInterface,":");   
      found=(int)strstr(line,curInterface);
      if (found){
	data=(char*)strchr(line,':');
	data++;	
	if(sscanf(data,"%i %i %i %i %i %i %i %i %i",&rxnew,&txnew,&collnew,&txnew,&txnew,&txnew,&txnew ,&txnew ,&txnew)<=0){
	  fclose(proc_net_dev);
	  errexit("ERROR: reading interface data \n");
	}
	/*
	print("%s:Rx=%d, Tx=%d, Col=%d\n",curInterface,rxnew, txnew, collnew);
	*/
	if (rxnew != last_net_results[ifnum][0]){
	  rxdiff+=rxnew-last_net_results[ifnum][0];	  
	  last_net_results[ifnum][0]=rxnew;
	}
	
	if (txnew != last_net_results[ifnum][1]){
	  txdiff+=txnew-last_net_results[ifnum][1];
	  last_net_results[ifnum][1]=txnew;
	}	
	ifnum++;
      	break;
      }
    }
  }

  fclose(proc_net_dev);

  /* If this is the first invocation to that function then return */
  if (firstnetcall==0){
    firstnetcall=1;
    return -1;
  }

  lastnetresult = ((100*(rxdiff+txdiff))/elapsedtime)/getMaxNet();
  /*print("NETWORK:current %d Bps, rxdiff=%d %d, txdiff=%d %d, maxnet=%d, elapsedtime=%d\n",(rxdiff+txdiff)/elapsedtime,rxdiff,rxnew,txdiff,txnew,getMaxNet(),elapsedtime);
   */
  return lastnetresult;  
}

/**
 * The following routine returns a number which 
 * indicates the percentage of disk space used.
 * @returns the percentage of the space used
 *          compared to the amount the user allowed,
 *          or -1 on error
 **/
int diskUsage() {
  DirName * contentdir = getContentDirectory();
  DIR * dinfo;
  int size;
  
  errno = 0;
  size = 0;
  dinfo = opendir((char*)contentdir);
  if ((errno == EACCES) || (dinfo == NULL))
    return -1;
  while (readdir(dinfo) != NULL) 
    size++; /* each file is approximately 1k (!) */
  closedir(dinfo);
  return (100*size)/getDiskQuota();
}

/**
 * The following routine returns a number between 0-100 which indicates
 * the percentage CPU usage.  * 
 * Before its first invocation the method initStatusCalls() must be called.
 * AFTER THAT THE METHOD SHOULD BE CALLED NO SOONER THAN 1 SECOND OR THE
 * FIRST TIME.
 * If there is an error the method returns -1
 **/
int cpuUsage(){
  static long lastcputime=0; /* In seconds */
  static int lastcpuresult=-1;
  static int firstcpucall=0;
  static int last_cpu_results[4] = { 0, 0, 0, 0 };

  FILE* proc_stat;
  char line[128];
  int user_read, system_read, nice_read, idle_read;
  int user, system, nice, idle;
  int ret=-1;
  int usage_time=0, total_time=1;

  struct timeval t1;
  long now, elapsedtime;

  /* Get the current time in microseconds */
  gettimeofday(&t1,NULL);
  now=t1.tv_sec;
  elapsedtime=now-lastcputime;
  lastcputime=now;
  
  proc_stat = fopen("/proc/stat", "r");
  /* Try to open the file*/
  if (! proc_stat) {
    print("ERROR: Could not open /proc/stat");
    return 0;
  }
  /* Get the first line with the data */
  if(fgets(line, 128, proc_stat)==NULL) {
    print("ERROR: reading /proc/stat file");
    fclose(proc_stat);
    return 0;
  }
  if(sscanf(line, "%*s %i %i %i %i", &user_read, &system_read, &nice_read,
	    &idle_read)<=0) {
    fclose(proc_stat);
    print("ERROR: decoding /proc/stat file");
    return 0;
  }

  /* Store the current usage*/
  user = user_read - last_cpu_results[0];
  system = system_read - last_cpu_results[1];
  nice = nice_read - last_cpu_results[2];
  idle = idle_read - last_cpu_results[3];
  
  /* Calculate the % usage */
  if ((user + system + nice + idle) > 0) {
    usage_time = user + system + nice;
    total_time = usage_time + idle;
  }
  ret= (100 * usage_time)/ total_time;

  /* Store the values for the next calculation*/
  last_cpu_results[0] = user_read;
  last_cpu_results[1] = system_read;
  last_cpu_results[2] = nice_read;
  last_cpu_results[3] = idle_read;
  fclose(proc_stat);

  /* If this is the first invocation to that function then return */
  if (firstcpucall==0){
    firstcpucall=1;    
    return -1;
  }
  /* If the method was called less that 1 second ago then return the old value
   */
  if (elapsedtime<=1){
    if (lastcpuresult >= 0)  
      return lastcpuresult;
    /* If no last time was ever obtained, retunr ERROR */
    else
      return -1;
  }
  lastcpuresult=ret;
  return ret;
}

/* end of statusCalls.c */
