/**
 * Hostkey - management of the node's files.
 * @author Christian Grothoff
 * @file primitives/storage.c
 **/

#include "storage.h"

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <errno.h>
#include <wordexp.h>

/**
 * Complete filename (a la shell) from abbrevition.
 * @param fil the name of the file, may contain ~/ or 
 *        other shell stuff. Will NOT be freed!
 * @returns the full file name, expanded with wordexp,
 *          NULL is returned on error
 **/
FileName expandFileName(FileName fil) {
  FileName fm;
  FileName fn;
  wordexp_t wordex;
  int res;

  if (fil == NULL)
    return NULL;
  res = wordexp(fil,&wordex,0);
  if (res != 0) {
    print("wordexp failed for %s, check your config file!\n",
	  fil);
    return NULL;
  }
  fm = wordex.we_wordv[0]; /* get first matching filename */
  if (fm == NULL) {
    print("wordexp failed returning NULL for %s, check your config file!\n",
	  fil);
    fn = malloc(strlen(fil)+1);
    memcpy(fn,fil,strlen(fil)+1);   
  } else {
    fn = malloc(strlen(fm)+1);
    memcpy(fn,fm,strlen(fm)+1);
  }
  wordfree(&wordex);
  return fn;
}

/**
 * implementation of "mkdir -p"
 * @param dir the directory to create
 **/
void mkdirp(DirName * dir) {
  FileName rdir;
  int len;
  int pos;

  rdir = expandFileName((FileName)dir); /* expand directory */
  len = strlen(rdir);
  pos = 0;
  while (pos < len) {
    if (rdir[pos] == '/') {
      rdir[pos] = 0;
      mkdir(rdir,S_IRUSR | S_IWUSR | S_IXUSR);      
      rdir[pos] = '/';       
    }      
    pos++;
  }   
  free(rdir);
}

/**
 * Build a filename from directory and filename, completing like the shell does
 * @param dir the name of the directory, may contain ~/ or other shell stuff. Will 
 *        NOT be freed!
 * @param fil the name of the file, will NOT be deallocated anymore!
 * @param result where to store the full file name (must be large enough!)
 **/
void buildFileName(DirName * dir,
		   HexName * fil,
		   FileName result) {
  if ((dir == NULL) || (fil == NULL) || (result == NULL))
    errexit("buildFileName called with NULL in arguments");
  result[0]=0;
  strcat(result,(char*)dir);
  strcat(result,(char*)fil);
}

/**
 * Read the contents of a binary file into a buffer.
 * @param fileName the name of the file, not freed,
 *        must already be expanded!
 * @param len the maximum number of bytes to read
 * @param result the buffer to write the result to
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readFile(FileName fileName,
	     int  len,
	     void * result) {
  /* open file, must exist, open read only */
  int handle;  
  int size;

  if ((fileName == NULL) || (result == NULL))
    return -1;
  handle = open(fileName,O_RDONLY,S_IRUSR);
  if (handle < 0)
    return -1; 
  size = read(handle, result, len);
  close(handle);
  return size;  
}

/**
 * Write a buffer to a file.
 * @param fileName the name of the file, NOT freed!
 * @param buffer the data to write
 * @param n number of bytes to write
 **/ 
void writeFile(FileName fileName, 
	       void * buffer,
	       int n) {
  int handle;
  /* open file, open with 600, create if not 
     present, otherwise overwrite */  
  if ((fileName == NULL) || (buffer == NULL))
    return;
  handle = open(fileName,O_CREAT|O_WRONLY,S_IRUSR|S_IWUSR);
  /* write the buffer take length from the beginning */
  if (n != write(handle,buffer,n)) 
    print("Writing %d bytes to file %s failed!\n",
	  n, fileName);
  close(handle);
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 * @param fn the name of the file (without directory)
 * @param block the CONTENT_SIZE bytes long block (no header!)
 **/
void writeContent(HexName * fn, 
		  CONTENT_Block * block) {
  CONTENT_Block preBlock[MAX_FSIZE];
  FileName fin;
  FileName ff; /* temporary */
  DirName * dir;
  int i;
  int j;
  int coll;
  int blocks;
  int handle;  

  dir = getContentDirectory();
  ff = malloc(strlen((char*)dir)+sizeof(HexName)+1);
  buildFileName(dir, fn, ff);
  fin = expandFileName(ff);
  free(ff);
  handle = open(fin, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);
  if (handle == -1) {
    print("Warning: failed to open file %s\n", fin);
    free(fin);
    return; /* failed! */
  }
  free(fin);
  j = read(handle, preBlock, MAX_FSIZE*CONTENT_SIZE);
  if ((j == 0)||(j % CONTENT_SIZE != 0)) {
    write(handle, block, CONTENT_SIZE); 
    close(handle);
    return;
  }
  blocks = j / CONTENT_SIZE;
  if (j == MAX_FSIZE) {
    print("Attention: MAX_FSIZE collisions reached for a keyword!\n");
    close(handle);
    return; /* full! */
  }
  /* collision! check if duplicate and if not, append */    
  coll = 0;
  for (i=0;i<blocks;i++) {
    for (j=0;j<CONTENT_SIZE;j++)
      if (preBlock[i].content[j] != block->content[j])
	coll++;
    if (coll == 0) {
      close(handle);
      return; /* collision! */
    }
  }
  write(handle, block, CONTENT_SIZE);
  close(handle);
}

/* how many files to read from the directory at once? 
   (we try to read N files, and if this failes, we try 2*n,
   and then 3*N, etc. ) */
#define DIR_FILE_COUNT 64

/**
 * Scan a directory for files. The name of the directory
 * must be expanded first (!).
 * @param dirName the name of the directory
 * @param result where to store the names of the files,
 *        only filenames with length 2*sizeof(HashCode160)
 *        will be stored.
 * @param max how many results should be obtained (max)
 * @return the number of files found, -1 on error
 **/
int scanDirectory(DirName * dirName,
		  DirContents * result,
		  int max) {
  DIR * dinfo;
  struct dirent *finfo;
  struct stat istat;
  int curPos;
 
  if ((dirName == NULL) || (result == NULL))
    return -1; 
  stat((char*)dirName,&istat);
  if (!S_ISDIR(istat.st_mode)) {
#ifdef PRINT_WARNINGS
    print("scanDirectory must be invoked on a directory (%s)!\n",
	  dirName);
#endif
    return -1;
  }
  errno = 0;
  dinfo = opendir((char*)dirName);
  if ((errno == EACCES) || (dinfo == NULL)) {
    print("scanDirectory: access denied (%s)\n",dirName);
    return -1;
  }
  curPos = 0;
  while ((finfo = readdir(dinfo)) != NULL)
    if (strlen(finfo->d_name) == sizeof(HashCode160)*2) {	
      memcpy(&result->files[curPos++].data[0],
	     finfo->d_name,
	     sizeof(HashCode160)*2+1);
      if (curPos == max) {
	closedir(dinfo);
	return curPos;
      }
    }
  closedir(dinfo);
  return curPos;
}

/**
 * Hash block of given size.
 * @param block the data to hash, length is given as a second argument
 * @param ret pointer to where to write the hashcode
 **/
void hash(void * block,
	  int size,
	  HashCode160 * ret) {
  if (block == NULL)
    return;
  RIPEMD160(block, size, (unsigned char*) ret);
}

static unsigned char * encoding__ = "0123456789ABCDEF";

/**
 * Convert (hash) block to hex (= filename)
 * @param block the sequence to convert
 * @param result where to store thestring (\0-terminated), hex-encoding
 **/
void hash2hex(HashCode160 * block,
	      HexName * result) {
  int i;
  int j;
  unsigned char c;
  unsigned char clow;

  if ((block == NULL) || (result == NULL)) 
    errexit("hash2hex called with block or result NULL!");

  result->data[sizeof(HashCode160)*2]=0;
  j=0;
  for (i=0;i<sizeof(HashCode160);i++) {
    c = ((unsigned char *)block)[i]; 
    clow = c & 15; /* get lower nibble */
    result->data[j++] = encoding__[clow];
    clow = c >> 4; /* get higher nibble */
    result->data[j++] = encoding__[clow];
  }  
}

/**
 * Convert hex (filename) to the hostIdentity
 * @param hex the filename
 * @param hash is set to the correspoinding host identity
 **/
void hex2hash(HexName * hex,
	      HashCode160 * hash) {
  int i;
  int j;
  unsigned char c;
  unsigned char clow;
  unsigned char chigh;

  if ((hex == NULL) || (hash == NULL))
    errexit("hex2hash called with hax or hash NULL!");
  assert(strlen((char*)hex) == sizeof(HashCode160)*2,
	 "assertion failed: strlen(hex) is not 32");
  j=0;
  i=0;
  while (i<sizeof(HashCode160)*2) {
    clow = hex->data[i++];
    if (clow >= 'A')
      clow = clow - 'A' + 10;
    else
      clow = clow - '0';
    chigh = hex->data[i++];    
    if (chigh >= 'A')
      chigh = chigh - 'A' + 10;
    else
      chigh = chigh - '0';
    c = clow + (chigh << 4);
    ((unsigned char *)hash)[j++] = c;
  }  
}

/**
 * Convert hex (filename) to the hostIdentity
 * @param hex the filename
 * @param ret address where to write the correspoinding host identity
 **/
void hex2hostIdentity(HexName * hex,
		      HostIdentity * ret) {
  hex2hash(hex,&ret->hashPubKey);
}

/* end of storage.c */
